# -*- coding: utf-8 -*-
from ._logging import get_logger
from ._options import get_options
from .annot2 import annotvcf
from . import makegenedata
from . import makedata
from . import convert
from . import precal
from . import preprocess
from . import web
from . import validate


def cli():
    global _log

    opt = get_options()
    opt.log = get_logger(silence=opt.silence, debug=opt.debug, logfile=opt.logfile)
    dispatch_job(opt)


def dispatch_job(opt):
    if opt.subcommand == 'annot' and opt.vcf != "":
        an = annotvcf.VCFAnnotator(opt)
        an.run()
        
    if opt.subcommand == 'makedata' and  opt.out != "":
        opt.vartype = opt.vartype.upper()
        if 'GENE' in opt.vartype:
            md = makegenedata.GeneDataSourceFile(opt)
            md.make_single_source_file()
        else:
            md = makedata.DataSourceGenerator(opt)
            if opt.check:
                md.check_datasourcefile()
            else:
                if opt.blocksize == '':
                    md.make_single_source_file()
                else:
                    md.make_single_source_file_with_block()
    if opt.subcommand == 'convert' and opt.infile != "":
        if opt.ds != "":
            cv = convert.CONVFILE(opt)
            cv.convert()
        if opt.vcf2tsv:
            cv = convert.VCF2TSV(opt)
            cv.convert()
        if opt.vep2tab:
            cv = convert.VEP2TAB(opt)
            cv.convert()
    if opt.subcommand == 'validate':
        if opt.vcf != "":
            va = validate.VCFValidator(opt)
            va.validate()

    if opt.subcommand == 'precal':
        mp = precal.PreCalculate(opt)
        mp.run()
    if opt.subcommand == 'preprocess':
        if opt.make_dbnsfp_transcript:
            obj = preprocess.MakeDbnsfpTranscript(opt.infile, opt.out, opt.ds, opt)
            obj.run()
    if opt.subcommand == 'web' and opt.ds != "":
        web.runserver(opt)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

SOURCELIST = ['ANNOVAR', 'SNPEFF', 'VEP']

DATAFILE = {}
DATAFILE['ANNOVAR'] = "ANNOT/b38_chrom/b38_WGSNV.annovar.chr#CHROM#.vcf.gz"
DATAFILE['SNPEFF'] = "ANNOT/b38_chrom/b38_WGSNV.snpeff.chr#CHROM#.vcf.gz"
DATAFILE['VEP'] = "ANNOT/b38_chrom/b38_WGSNV.vep.chr#CHROM#.vcf.gz"

ANNOVAR = ""
SNPEFF = ""
VEP = ""

if __name__ == "__main__":
    pass



import os
import json
from .util import file_util
from .util import struct_util
from .util import seq_util
from . import external_functions


RESERVED_COL = ["chrom", "spos", "epos", "ensgid"]


def is_available(field):
    flag = True
    if struct_util.get_dict_value(field, "is_available", True) == False:
        flag = False
    if struct_util.get_dict_value(field, "do_import", True) == False:
        flag = False
    return flag


def is_reserved_column(f1):
    flag = False
    if f1['name'] in RESERVED_COL or ('name2' in f1.keys() and f1['name2'] in RESERVED_COL):
        flag = True
    return flag


def strip_value(v1):
    v1 = v1.replace('"', '')
    return v1


def encode_value(v1, delimiter=""):
    if v1 == '.' or v1 == '-':
        v1 = ''
    if delimiter != "":
        if delimiter != '|':
            v1 = v1.replace("|", "%7C")
            v1 = v1.replace(delimiter, '|')
    # v1 = urllib.parse.quote(v1)
    return v1


class GeneSourceReader():
    def __init__(self, datastruct, datafile_path):
        self.datastruct = datastruct
        self.source_name = self.datastruct['name']
        self.fname = os.path.join(datafile_path, self.datastruct['datafile'])
        self.fileformat = self.datastruct['format']
        self.target_colidx = []
        self.header = []
        self.subembedded = struct_util.get_dict_value(datastruct, 'subembedded', '')
        self.delimiter = {}
        self.list_type_fields = {}
        self.defaultvalue = {}
        self.field_function = {}
        self.field_function_param = {}
        self.field_names = []
        self.key_colidx = -999
        self.reserved_colidx = {}
        self.reserved_data = {}
        self.field_type = {}
        self.data = {}
        self.ensgid_list = []
        self.is_range_data = False
        self.target_colnames = []
        self.target_colnames2 = []
        self.is_ensemblegene = struct_util.get_dict_value(datastruct, 'is_ensemblegene', False)
        self.read_header()
        print("loading..", self.source_name)
        # print(self.header)
        self.set_target_column(self.datastruct['fields'])

        if self.key_colidx == -999:
            self.load_range_data()
            self.is_range_data = True
        else:
            self.load_data()

    def split_into_fields(self, line):
        arr = line.split('\t')
        arr[-1] = arr[-1].strip()
        return arr

    def split_into_fields2(self, line):
        splited_fields = []
        for field1 in line.split('\t'):
            for f1 in field1.strip().split('|'):
                splited_fields.append(f1)
        return splited_fields

    def read_header(self):
        for line in file_util.gzopen(self.fname):
            if self.fname.endswith('.gz'):
                line = line.decode('UTF-8')
            if line[0] == '#' and line[1] != '#':
                self.header = self.split_into_fields(line[1:])
                break

    def set_target_column(self, fields_structure):
        name2_map = {}
        for f1 in fields_structure:

            if is_available(f1):
                self.field_names.append(f1['name'])

                # print("self.header",self.header, f1['name'], self.key_colidx)
                colidx = self.header.index(f1['name'])
                if f1['name'] in self.header:
                    if is_reserved_column(f1):
                        for resv in RESERVED_COL:
                            if f1['name'] == resv or ('name2' in f1.keys() and f1['name2'] == resv):
                                self.reserved_colidx[resv] = self.header.index(f1['name'])
                    else:
                        if 'is_list' in f1.keys() and f1['is_list']:
                            self.list_type_fields[self.header.index(f1['name'])] = f1['is_list']
                        if 'delimiter' in f1.keys():
                            self.delimiter[self.header.index(f1['name'])] = f1['delimiter']
                        if 'default' in f1.keys():
                            self.defaultvalue[self.header.index(f1['name'])] = f1['default']

                        self.target_colidx.append(colidx)
                else:
                    if not is_reserved_column(f1):
                        self.target_colidx.append(-999)

                if 'type' in f1.keys():
                    self.field_type[colidx] = f1['type']

                if 'function' in f1.keys() and f1['function'] != '':
                    self.field_function[colidx] = f1['function']
                    self.field_function_param[colidx] = ''
                    if 'param' in f1.keys() and f1['param'] != '':
                        self.field_function_param[colidx] = f1['param']

                name2_map[colidx] = struct_util.get_dict_value(f1, 'name2', '')


        if 'ensgid' in self.reserved_colidx.keys():
            self.key_colidx = self.reserved_colidx['ensgid']

        self.target_colnames = []
        self.target_colnames2 = []
        for tidx in self.target_colidx:
            self.target_colnames.append(self.header[tidx])
            if name2_map[tidx] == "":
                self.target_colnames2.append(self.header[tidx])
            else:
                self.target_colnames2.append(name2_map[tidx])


    def rm_version_in_ensgid(self, ensg):
        return ensg.strip().split('.')[0]


    def get_value_using_eval_field_function(self, colidx, cidx_no, cidxvalue):
        exec_str = "external_functions." + self.field_function[colidx]
        exec_str += '('
        paramstr_list = []
        for param in self.field_function_param[colidx].split(','):
            if param == "mutanno_value_variantkey":
                pstr = 'variantkey'
            elif cidx_no == self.field_names.index(param):
                pstr = 'cidxvalue'
            else:
                pstr = 'arr_selected_fields['
                pstr += str(self.field_names.index(param))
                pstr += ']'
            paramstr_list.append(pstr)
        exec_str += ','.join(paramstr_list)
        exec_str += ')'
        cidxvalue = eval(exec_str)
        return cidxvalue

    def typecast(self, value, value_type):
        if (value == '' or value == 'NA') and value_type != 'string':
            rst_value = None
        else:
            if value_type == 'integer':
                try:
                    rst_value = int(value)
                except ValueError:
                    print('ValueError:',value)
                rst_value = int(value)
            elif value_type == 'number':
                try:
                    rst_value = float(value)
                except ValueError:
                    print('ValueError:',value)
                rst_value = float(value)
            elif value_type == 'boolean':
                rst_value = boolean(value)
            else:
                rst_value = value
        return rst_value

    def load_data(self):
        for line in file_util.gzopen(self.fname):
            if self.fname.endswith('.gz'):
                line = line.decode('UTF-8')
            if line[0] != '#':
                arr = self.split_into_fields(line)

                ensgid = self.rm_version_in_ensgid(arr[self.key_colidx])
                self.ensgid_list.append(ensgid)
                try:
                    self.data[ensgid]
                except KeyError:
                    self.data[ensgid] = []

                cidx_no = 0
                target_data = []
                for colidx in self.target_colidx:
                    cidx_no += 1
                    if colidx == -999:
                        cidxvalue = ""
                    else:
                        cidxvalue = strip_value(arr[colidx])

                    if colidx in self.field_function.keys():
                        cidxvalue = self.get_value_using_eval_field_function(colidx, cidx_no, cidxvalue)
                        
                    delimiter = ''
                    if colidx in self.delimiter.keys():
                        delimiter = self.delimiter[colidx]
                    if colidx in self.list_type_fields.keys():
                        cidxvaluelist = []
                        for v1 in cidxvalue.split(delimiter):
                            cidxvaluelist.append(self.typecast(v1, self.field_type[colidx]))
                        target_data.append(cidxvaluelist)
                    else:
                        target_data.append(self.typecast(cidxvalue, self.field_type[colidx]))


                self.data[ensgid].append(target_data)

                for resv in self.reserved_colidx.keys():
                    colidx = self.reserved_colidx[resv]
                    try:
                        self.reserved_data[resv]
                    except KeyError:
                        self.reserved_data[resv] = {}

                    try:
                        self.reserved_data[resv][ensgid]
                    except KeyError:
                        if resv == 'ensgid':
                            self.reserved_data[resv][ensgid] = self.rm_version_in_ensgid(arr[colidx])
                        else:
                            self.reserved_data[resv][ensgid] = self.typecast(strip_value(arr[colidx])
                                , self.field_type[colidx])

    def load_range_data(self):
        for line in file_util.gzopen(self.fname):
            line = line.decode('UTF-8')
            if line[0] != '#':
                arr = line.split('\t')
                arr[-1] = arr[-1].strip()
                chrom = arr[0].strip()
                spos = int(arr[1].strip())
                epos = int(arr[2].strip())
                d = {'spos': spos, 'epos': epos, 'v': []}
                for cidx in self.target_colidx:
                    if cidx == -999:
                        d['v'].append('')
                    else:
                        d['v'].append(strip_value(arr[cidx]))
                    try:
                        self.data[chrom].append(d)
                    except KeyError:
                        self.data[chrom] = [d]

    def get_range_data(self, geneinfo):
        rst = []
        geneinfo['epos'] = int(geneinfo['epos'])
        geneinfo['spos'] = int(geneinfo['spos'])
        try:
            for v1 in self.data[geneinfo['chrom']]:
                if geneinfo['epos'] >= v1['spos'] and geneinfo['spos'] <= v1['epos']:
                    rst.append(v1['v'])
                elif geneinfo['epos'] < v1['spos']:
                    break
        except KeyError:
            pass
        return rst

    def is_empty_value(self, v1):
        flag = False
        if v1 == "":
            flag = True
        if type(v1) == list:
            if len(v1) == 0:
                flag = True
            if len(v1) == 1 and v1[0] == "":
                flag = True
        return flag

    def get_data_dict_ensgid(self, ensgid, geneinfo, colname_type='name'):
        rstdata = []
        if self.is_range_data:
            datalist = self.get_range_data(geneinfo)
        else:
            try:
                datalist = self.data[ensgid]
            except KeyError:
                datalist = []
    
        if self.is_range_data and len(self.target_colnames) == 1:
            # for special case (CYTOBAND)
            d = []
            for datarow in datalist:
                d.append(datarow[0])
            if not self.is_empty_value(d):
                rstdata.append({self.target_colnames[0]:d})
        else:
            for datarow in datalist:
                d = {}
                for k in range(len(self.target_colnames)):
                    if colname_type == 'name2':
                        k1 = self.target_colnames2[k]
                    else:
                        k1 = self.target_colnames[k]
                    if not self.is_empty_value(datarow[k]) and datarow[k] != None:
                        d[k1] = datarow[k]
                rstdata.append(d)

        return rstdata


class GeneSourceMerger():
    def __init__(self, source_readers):
        self.source_readers = source_readers
        self.ensgid_list = []
        self.colnames = []
        self.colnames2 = []
        self.set_colnames()
        self.merge_ensgid_list()

    def merge_ensgid_list(self):
        ensgid_map = {}
        for sid in self.source_readers.keys():
            s1 = self.source_readers[sid]
            # print(len(s1.ensgid_list), s1.ensgid_list[:3])
            for ensgid in s1.ensgid_list:
                if ensgid != '':
                    ensgid_map[ensgid] = 1
        self.ensgid_list = list(ensgid_map.keys())

    

    def set_colnames(self):
        self.colnames = {}
        self.colnames2 = {}
        for sid in self.source_readers.keys():
            s1 = self.source_readers[sid]
            self.colnames[sid] = s1.target_colnames
            self.colnames2[sid] = s1.target_colnames2

    def get_data_dict(self, ensgid):
        data = {}
        resv = self.get_reserved_data(ensgid)

        data = {}
        if 'spos' in resv.keys() and 'epos' in resv.keys():
            for r1 in RESERVED_COL:
                try:
                    v1 = resv[r1]
                except KeyError:
                    v1 = ''
                data[r1] = v1


            for sid in self.source_readers.keys():
                s1 = self.source_readers[sid]

                d1 = s1.get_data_dict_ensgid(ensgid, resv, 'name2')
                if s1.subembedded == "":
                    # if len(d1) > 2:
                    #     print(d1)
                    if len(d1) > 0:
                        data.update(d1[0])
                else:
                    if len(d1) > 0:
                        data.update({s1.subembedded:d1})
        return data

    def get_reserved_data(self, ensgid):
        resv = {}
        for sid in self.source_readers.keys():
            s1 = self.source_readers[sid]
            if s1.is_ensemblegene:
                for r1 in s1.reserved_data.keys():
                    try:
                        resv[r1] = s1.reserved_data[r1][ensgid]
                    except KeyError:
                        pass
        return resv

    def get_data_list(self, ensgid):
        data = []
        resv = self.get_reserved_data(ensgid)

        if 'spos' in resv.keys() and 'epos' in resv.keys():
            for r1 in RESERVED_COL:
                try:
                    r2 = resv[r1]
                except KeyError:
                    r2 = ''
                data.append(r2)

            for sid in self.source_readers.keys():
                s1 = self.source_readers[sid]
                if s1.is_range_data:
                    rdata = self.get_range_data(s1.data, resv)
                    if len(rdata) > 0:
                        data.extend(rdata[0])
                    else:
                        for cidx in s1.target_colidx:
                            data.append('')
                else:
                    try:
                        data.extend(s1.data[ensgid])
                    except KeyError:
                        for cidx in s1.target_colidx:
                            data.append('')
        print('get_data_list',data)
        return data


class GeneDataSourceFile():
    def __init__(self, opt):
        self.vartype = opt['vartype']
        self.outtype = opt['outtype']
        self.set_outfile_extension(opt['out'], opt['outtype'])
        self.datastruct = file_util.load_json(opt['ds'])
        self.region = opt['region']
        self.blocksize = opt['blocksize']
        

    def set_outfile_extension(self, out, outtype):
        out2 = out.replace('.'+outtype+'.gz', '.' + outtype)
        if out2[-4:] != "." + outtype:
            out2 += "." + outtype
        self.out = out2

    def get_bed_header(self):
        header = []
        for r1 in RESERVED_COL:
            header.append(r1)
        info_arr = []
        for s1 in self.datastruct['gene_source']:
            if is_available(s1):
                subembedded = struct_util.get_dict_value(s1, "subembedded", "")

                info_header = []
                for f1 in s1['fields']:
                    if is_available(f1):
                        if 'name2' in f1.keys() and f1['name2'] != '':
                            fname = f1['name2']
                        else:
                            fname = f1['name']
                        if fname not in header:
                            info_header.append(s1['name'] + "_" + fname)
                info_arr.append('|'.join(info_header))
        header.append('\t'.join(info_arr))
        return '#' + '\t'.join(header)

    def save_json_out(self):
        # with open(self.out, 'w') as fp:
        fp = open(self.out, 'w')
        json.dump(data, fp)


    def save_file_header(self, fp):
        if self.outtype == "json":
            fp.write('[')

    def save_file_tail(self, fp):
        if self.outtype == "json":
            fp.write(']')

    def save_file_record(self, fp, data, prev_flag):
        
        if self.outtype == "json":
            flag = True
            if "CODING_GENE" in self.vartype:
                if data['gene_biotype'] not in ["protein_coding", "miRNA", "polymorphic_pseudogene"]:
                    flag = False
            if "MAIN_CHROM" in self.vartype:
                if data['chrom'] not in seq_util.MAIN_CHROM_LIST:
                    flag = False

            if flag:
                if prev_flag:
                    fp.write(',')
                # fp.write(json.dumps(data))
                fp.write(json.dumps(data, indent=2))
        else:
            fp.write(data)
        return flag


    # init function
    def make_single_source_file(self):
        file_util.check_dir(self.out)        
        fp = open(self.out, 'w')
        # fp.write(self.get_bed_header() + '\n')

        sidlist = []
        source_readers = {}
        for s1 in self.datastruct['gene_source']:
            if is_available(s1):
                sid = s1['name']
                sidlist.append(sid)

                datafile_path = ''
                if 'datafile_path' in self.datastruct.keys():
                    datafile_path = self.datastruct['datafile_path']
                source_readers[sid] = GeneSourceReader(s1, datafile_path)

        gsm = GeneSourceMerger(source_readers)

        i = 0
        self.save_file_header(fp)
        flag_write = False
        for ensgid in gsm.ensgid_list:
            # data = gsm.get_data_list(ensgid)
            # line = '\t'.join(data)
            data = gsm.get_data_dict(ensgid)
            if len(data.keys()) > 0:
                flag2 = self.save_file_record(fp, data, flag_write)
                if flag2:
                    flag_write = True
                i += 1

            if i > 1000:
                pass
                # break
        self.save_file_tail(fp)
        print('Saved', self.out)
        fp.close()

        ### file check
        print('Json checking...', self.out)
        fp = open(self.out)
        data = json.load(fp)






#!/usr/bin/env python
# -*- coding: utf-8 -*-
# file_util.py
import os
import sys
import tabix
import json
from . import proc_util
import pprint

def print(str):
    pp = pprint.PrettyPrinter(indent=2)
    pp.pprint(str)

def save_gzip(out):
    cmd = "bgzip -c " + out + " > " + out + ".gz"
    proc_util.run_cmd(cmd)

def strip_gzext(fname):
    rst = fname
    if rst.endswith('.gz'):
        rst = rst[:-3]
    return rst

def line2arr(line, delimiter = '\t'):
    arr = line.split('\t')
    arr[-1] = arr[-1].strip()
    return arr


def zero_format(i, size):
    # 00001000 format
    s1 = "{:0>" + str(size) + "d}"
    return s1.format(i)


def load_json(jsonfile):
    ds = ""
    with open(jsonfile) as jfp:
        ds = json.load(jfp)
    return ds


def get_tblist(vcfgzlist):
    tblist = []
    for vcfgz in vcfgzlist:
        tblist.append(tabix.open(vcfgz))
    return tblist


def getRowTabix(tb, chrom, pos, ref="", alt=""):
    if not isinstance(pos, int):
        pos = int(pos)
    recs = tb.query(chrom, pos - 1, pos)
    rec = []
    for r1 in recs:
        if int(r1[1]) == pos:
            if ref != "" and alt != "":
                if r1[3] == ref and r1[4].replace(",<NON_REF>", "") == alt:
                    rec = r1
                    break
            elif ref != "":
                if r1[3] == ref:
                    rec = r1
                    break
            else:
                rec = r1
                break

    return rec


def check_dir(fname):
    if fname[0] != '/' and fname[0] != '.':
        fname = './' + fname
    arr = fname.split("/")
    fpath = arr[0]
    for d in arr[1:-1]:
        fpath += "/" + d
        if not is_exist(fpath):
            mkDir(fpath)
            # print("make directory : " + fpath)


def fileOpenOffset(filepath, start_pos, end_pos):
    f = open(filepath)
    f.seek(start_pos)
    cont = f.read(end_pos - start_pos)
    return cont


def fileOpen2(path, opt):
    try:
        thisFile = open(path, opt)
    except IOError:
        print(sys.stderr, "File could not be opened")
        sys.exit(1)
    else:
        return thisFile


def fileOpen(path):
    cont = ""
    if path.endswith('.gz'):
        for line in gzopen(path):
            cont += decodeb(line)
    else:
        f = open(path, "r")
        cont = f.read()
    return cont


def fileSave(path, cont, opt, gzip_flag="n"):
    if gzip_flag == "gz":
        import gzip
        f = gzip.open(path, opt)
        f.write(cont)
        f.close()
    else:
        f = open(path, opt)
        f.write(cont)
        f.close


def gzsave(path, cont, opt):
    return fileSave(path, cont, opt, "gz")


def listdir(dirPath, ext=""):
    flist = []
    for fname in os.listdir(dirPath):
        if (len(ext) > 0 and fname.endswith(ext)) or len(ext) == 0:
            flist.append(fname)
    flist.sort()
    return flist


def walk(dirPath, ext=""):
    flist = []
    for root, dirs, files in os.walk(dirPath):
        for fname in files:
            if (len(ext) > 0 and fname.endswith(ext)) or len(ext) == 0:
                fullpath = os.path.join(root, fname)
                flist.append(fullpath)
    flist.sort()
    return flist


def walkdir(dirPath, cls="simple"):
    dlist = []
    for root, dirs, files in os.walk(dirPath):
        for dname in dirs:
            if dname != "":
                if cls == "simple":
                    dlist.append(dname)
                else:
                    dlist.append(os.path.join(root, dname))
    dlist.sort()
    return dlist


def read_table_col(filename, delemeter="\t"):
    tab = {}
    i = 0
    header = []
    repeatmap = {}
    for line in gzopen(filename):
        if len(line.strip()) > 0:
            arr = line.split(delemeter)
            arr[-1] = arr[-1].strip()

            if i == 0:
                for j in range(0, len(arr)):
                    h1 = arr[j].strip()
                    if h1 in header:
                        header.append(h1 + "_" + str(repeatmap[h1]))
                        repeatmap[h1] += 1
                    else:
                        header.append(h1)
                        repeatmap[h1] = 1
            else:
                for j in range(0, len(header)):
                    try:
                        tab[header[j]].append(arr[j].strip())
                    except KeyError:
                        tab[header[j]] = []
                        tab[header[j]].append(arr[j].strip())
        i += 1
    return tab


def read_table(filename, delemeter="\t"):
    tab = []
    i = 0
    header = []
    repeatmap = {}
    for line in gzopen(filename):
        if filename.endswith('.gz'):
            line = line.decode('UTF-8')
        if len(line.strip()) > 0:
            arr = line.split(delemeter)
            arr[-1] = arr[-1].strip()

            if i == 0:
                for j in range(0, len(arr)):
                    h1 = arr[j].strip()
                    if h1 in header:
                        header.append(h1 + "_" + str(repeatmap[h1]))
                        repeatmap[h1] += 1
                    else:
                        header.append(h1)
                        repeatmap[h1] = 1
            else:
                t1 = {}
                for j in range(0, len(header)):
                    t1[header[j]] = arr[j].strip()
                tab.append(t1)
        i += 1
    return tab


def read_table_key(filename, key_col_idx=0, delemeter="\t"):
    tab = {}
    i = 0
    header = []
    repeatmap = {}
    for line in open(filename):
        if len(line.strip()) > 0:
            arr = line.split(delemeter)
            arr[-1] = arr[-1].strip()

            if i == 0:
                for j in range(0, len(arr)):
                    h1 = arr[j].strip()
                    if h1 in header:
                        header.append(h1 + "_" + str(repeatmap[h1]))
                        repeatmap[h1] += 1
                    else:
                        header.append(h1)
                        repeatmap[h1] = 1
            else:
                t1 = {}
                for j in range(0, len(header)):
                    t1[header[j]] = arr[j].strip()
                tab[arr[key_col_idx]] = t1
        i += 1
    return tab


def comp_gz(target_file, gzip_file=""):
    import gzip
    if gzip_file == "":
        gzip_file = target_file + ".gz"
    f_in = open(target_file, 'rb')
    f_out = gzip.open(gzip_file, 'wb')
    f_out.writelines(f_in)
    f_out.close()
    f_in.close()


def mkDir(dpath):
    if not os.path.isdir(dpath):
        os.mkdir(dpath)


def is_exist(fpath):
    return os.path.exists(fpath)


def getFileInfo(fpath):
    st = {}
    (st['mode'], st['ino'], st['dev'], st['st_nlink'], st['uid'], st['gid'],
     st['size'], st['atime'], st['mtime'], st['ctime']) = os.stat(fpath)
    return st


def getFileSize(fpath):
    st = {}
    (st['mode'], st['ino'], st['dev'], st['st_nlink'], st['uid'], st['gid'],
     st['size'], st['atime'], st['mtime'], st['ctime']) = os.stat(fpath)
    return st['size']


def copy(src, dst):
    import shutil
    shutil.copy(src, dst)


def rm(fname):
    if is_exist(fname):
        os.remove(fname)


def fopen(fname):
    if fname.endswith(".gz"):
        import gzip
        f1 = gzip.GzipFile(fname, "r")
    else:
        f1 = open(fname)
    return f1


def gzopen(fname):
    if fname.endswith(".gz"):
        import gzip
        f1 = gzip.GzipFile(fname, "r")
    else:
        f1 = open(fname)
    return f1


def extract_line(fname, findme, column=-9, delimiter="space", outfile="", header=False):
    import str_util
    if fname[-3:] == ".gz":
        f1 = gzopen(fname)
    else:
        f1 = open(fname)

    if outfile != "":
        fileSave(outfile, "", "w")

    i = 0
    for line in f1:
        if header and i == 0:
            cont = line.strip()
        else:
            cont = ""
            if column < 0:
                if findme in line:
                    cont = line.strip()
            else:
                if delimiter == "space":
                    line = str_util.strip_space(line)
                    arr = line.split(" ")
                else:
                    arr = line.strip().split(delimiter)
                if findme in arr[column]:
                    cont = line.strip()

        if cont != "" and outfile != "":
            fileSave(outfile, cont + "\n", "a")
        elif cont != "":
            print(cont)

        i += 1


# Function Name : load_seies_file
# load data file with series format in GEO database
def load_series_file(fname, coat="n"):
    m = {}
    cont = ""
    for line in gzopen(fname):
        if line[0] == "!":
            arr = line[1:].strip().split('"\t"')
            arr_k = arr[0].split("\t")
            k = arr_k[0].strip()
            if len(arr_k) > 1:
                v = [arr_k[1].strip()]
            else:
                v = []
            if len(arr) > 1:
                v.extend(arr[1:])
            for i in range(len(v)):
                v[i] = v[i].replace('"', '')
            if len(v) > 0:
                try:
                    arr2 = m[k]
                    arr2.append(v)
                    '''
                    if len(arr) == 2:
                        arr2.append(v)
                    else:
                        arr2.append(v)
                    '''
                except KeyError:
                    m[k] = []
                    m[k].append(v)
                    '''
                    if len(arr) == 2:
                        m[k].append(v)
                    else:
                        m[k].append(v)
                    '''
            else:
                if len(cont) > 0:
                    k = k.replace("_end", "")
                    try:
                        arr2 = m[k]
                        arr2.append(cont)
                        m[k] = arr2
                    except KeyError:
                        m[k] = []
                        m[k].append(cont)
                    cont = ""
        else:
            cont += line
    return m


def log(msg, logfile="", silence=False):
    import time_util

    if "#DATE#" in logfile:
        logfile = logfile.replace("#DATE#", time_util.getToday())

    cont = "[" + time_util.getNow() + "] " + msg + "\n"

    if not silence:
        print(cont,)

    if logfile != "":
        fileSave(logfile, cont, "a")


def get_header(fname):
    arr = []
    for line in gzopen(fname):
        if fname.endswith(".gz"):
            line = line.decode('UTF-8')
        arr = line.split('\t')
        arr[-1] = arr[-1].strip()
        break
    return arr


def get_header_idx(fname):
    h = {}
    arr = get_header(fname)
    for k in range(len(arr)):
        h[arr[k]] = k
    return h


def decodeb(bstr):
    if isinstance(bstr, bytes):
        rst = bstr.decode('UTF-8')
    else:
        rst = bstr
    return rst


def rmext(fname, ext):
    return fname[:(-1 * len(ext))]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

VCF_COL = ['CHROM','POS','ID','REF','ALT','QUAL','FILTER','INFO','FORMAT']

INFOIDX = VCF_COL.index('INFO')

# https://www.ncbi.nlm.nih.gov/assembly/GCF_000001405.39
CHROM_HGVS = {}
CHROM_HGVS["1"] = "NC_000001.11"
CHROM_HGVS["2"] = "NC_000002.12"
CHROM_HGVS["3"] = "NC_000003.12"
CHROM_HGVS["4"] = "NC_000004.12"
CHROM_HGVS["5"] = "NC_000005.10"
CHROM_HGVS["6"] = "NC_000006.12"
CHROM_HGVS["7"] = "NC_000007.14"
CHROM_HGVS["8"] = "NC_000008.11"
CHROM_HGVS["9"] = "NC_000009.12"
CHROM_HGVS["10"] = "NC_000010.11"
CHROM_HGVS["11"] = "NC_000011.10"
CHROM_HGVS["12"] = "NC_000012.12"
CHROM_HGVS["13"] = "NC_000013.11"
CHROM_HGVS["14"] = "NC_000014.9"
CHROM_HGVS["15"] = "NC_000015.10"
CHROM_HGVS["16"] = "NC_000016.10"
CHROM_HGVS["17"] = "NC_000017.11"
CHROM_HGVS["18"] = "NC_000018.10"
CHROM_HGVS["19"] = "NC_000019.10"
CHROM_HGVS["20"] = "NC_000020.11"
CHROM_HGVS["21"] = "NC_000021.9"
CHROM_HGVS["22"] = "NC_000022.11"
CHROM_HGVS["X"] = "NC_000023.11"
CHROM_HGVS["Y"] = "NC_000024.10"


def get_hgvsg(chrom, pos, ref, alt):
    try:
        hgvsg = CHROM_HGVS[chrom.replace('chr','')] + ':g.'
        if len(ref) > len(alt):
            if len(ref) == 2:
                hgvsg += str(pos + 1) + 'del'
            else:
                hgvsg += str(pos + 1) + '_' + str(pos + len(ref) - 1) + 'del'
        elif len(ref) < len(alt):
            hgvsg += str(pos) + '_' + str(pos + 1) + 'ins' + alt[1:]
        else:
            hgvsg += str(pos) + ref + '>' + alt
    except KeyError:
        hgvsg = ""
    return hgvsg


def get_variant_class(ref, alt):
    # TODO: need to upgrade for ',' in alt.
    vcls = ""
    if len(ref) == len(alt) and len(alt) == 1:
        vcls = "SNV"
    elif len(ref) < len(alt):
        vcls = "INS"
    elif len(ref) > len(alt):
        vcls = "DEL"
    return vcls

def add_info(info1, info2):
    info1 = strip_info(info1)
    info2 = strip_info(info2)
    if info1 == "":
        info = info2
    if info2 == "":
        info = info1
    if info1 != "" and info2 != "":
        info = info1 + ';' + info2
    return info

def strip_info(info):
    if info == ".":
        info = ""
    if info != "" and info[-1] == ";":
        info = info[:-1]
    return info

def convert_to_metadata(d):
    cont = []
    for field in d.keys():
        if type(d[field]) == type({}):
            for key in d[field].keys():
                valuearr = ["ID=" + key]
                for k2 in d[field][key].keys():
                    if k2 in ['Number','Type']:
                        valuearr.append(k2 + '=' + d[field][key][k2])
                    else:
                        valuearr.append(k2 + '="' + d[field][key][k2] + '"')
                value = '<' + ','.join(valuearr) + '>'
                cont.append('##'+field + '=' + value)
        else:
            value = d[field]
            cont.append('##'+field + '=' + value)
    return '\n'.join(cont)

def get_info_header(headertype, infoid, version, vdate,  infodesc, subfields, sourcesubembed=""):
    header = ""

    if headertype == "MUTANNO":
        header += "##MUTANNO=<ID="+infoid+",Version=\""+version+"\",Date=\""+vdate+"\">\n"
    elif headertype == "INFO":
        header += "##INFO=<ID=" + infoid + ",Number=.,Type=String"
        header += ",Description=\"" + infodesc + ". "
        if sourcesubembed != '':
            header += "Subembedded:'" + sourcesubembed + "':"
        header += "Format:'" + '|'.join(subfields) + "'\">\n"
    return header


# numgt: number format of genotype (0/0, 1/1)
# ref: reference allele
# alt: alternative allele or allels
def get_genotype(numgt, ref, alt):
    numgt = numgt.replace('|', '/')
    allele = [ref]
    allele.extend(alt.split(','))
    gt = []
    for ngt in numgt.split('/'):
        if ngt == '.':
            gt.append('.')
        else:
            gt.append(allele[int(ngt)])
    return '/'.join(gt)


def split_multiallelic_variants(vcfrecord):
    arralt = vcfrecord[4].split(',')
    rst = []
    for k in range(len(arralt)):
        r1 = []
        for j in range(len(vcfrecord)):
            if j >= 9:
                arr = vcfrecord[j].split(':')
                # GT
                phase_hipen = arr[0][1]
                if arr[0] != '0/0' and arr[0] != '0|0':
                    gt = arr[0].split('/')
                    if str(k+1) in gt:
                        if "0" in gt:
                            arr[0] = "0" + phase_hipen + "1"
                        else:
                            arr[0] = "1" + phase_hipen + "1"
                    else:
                        arr[0] = "0" + phase_hipen + "0"
                # AD
                ad = arr[1].split(',')
                arr[1] = ad[0] + ',' + ad[k+1]
                r1.append(':'.join(arr))
            else:
                r1.append(vcfrecord[j])
        r1[4] = arralt[k]

        # info
        info = []
        for f1 in r1[7].split(';'):
            if "=" in f1:
                arr = f1.split('=')
                if arr[0] in ['AC','AF','MLEAC','MLEAF']:
                    arr2 = arr[1].split(',')
                    f1 = arr[0] + "=" + arr2[k]
            info.append(f1)
        # info.append('multiallele=' + vcfrecord[0] + ':' + vcfrecord[1] + '%20' + vcfrecord[3] + '/' + vcfrecord[4])
        r1[7] = ';'.join(info)
        rst.append(r1)
    return rst


def encode_value(v1):
    v1 = v1.replace(';', "%3B").replace('=', "%3D").replace("|", "%7C").replace(",", "%2C")
    v1 = v1.replace('"', "%22").replace('~', "%7E").replace(' ', '%20')
    return v1


def encode_infovalue(v1, delimiter=""):
    if v1 is None:
        rst = ""
    elif isinstance(v1, bool):
        if v1:
            rst = "1"
        else:
            rst = "0"
    else:
        rst = str(v1)
        if rst == '.' or rst == '-':
            rst = ''
        rst = encode_value(rst)
        if delimiter != "":
            rst = rst.replace(encode_value(delimiter), '~')
            
    return rst


def remove_nonvariant_base(ref, alt, pos=0, miss_char=''):
    p = 0
    if min(len(ref), len(alt)) >= 1:
        for j in range(min(len(ref), len(alt))):
            if len(ref) == 0 or len(alt) == 0:
                break
            if alt[0] == ref[0]:
                alt = alt[1:]
                ref = ref[1:]
                p += 1
            else:
                break
        for i in range(1, min(len(ref), len(alt))):
            if len(ref) == 0 or len(alt) == 0:
                break
            if alt[-1] == ref[-1]:
                alt = alt[:-1]
                ref = ref[:-1]
            else:
                break
    if ref == '':
        ref = miss_char
    if alt == '':
        alt = miss_char
    pos += p
    return (ref, alt, pos)


class VCFHEADER():
    collist = []
    mutannofield = {}
    header = ""

    def __init__(self, header):
        self.header = header
        self.pars()

    def pars(self):
        flag_mutanno = False
        m = {}
        for line in self.header.split('\n'):
            if line[:2] == "##":
                if line[:len("##MUTANNO=<ID=")] == "##MUTANNO=<ID=":
                    arr = line.replace('##MUTANNO=<ID=', '').replace('>', '').split(',')
                    d = {}
                    d['sourcename'] = arr[0].strip()
                    for f1 in arr[1:]:
                        arr2 = f1.strip().replace('""', '').split('=')
                        d[arr2[0].strip()] = arr2[1].strip()
                    d['fieldlist'] = []
                    m[arr[0].strip()] = d

                if line[:len("##MUTANNO=<ID=MUTANNO")] == "##MUTANNO=<ID=MUTANNO":
                    flag_mutanno = True

                if line[:len("##INFO=")] == "##INFO=" and 'Format:' in line and flag_mutanno:
                    arr = line.strip().split(',')
                    sourcename = arr[0].replace('##INFO=<ID=', '')
                    if sourcename in m.keys():
                        m[sourcename]['fieldlist'] = line.split(
                            'Format:')[-1].replace("'", "").replace('">', '').strip().split('|')
            elif line[:len("#CHROM")] == "#CHROM":
                self.collist = line[1:].strip().split('\t')
        self.mutannofield = m


#     fieldkeynamelist.append( conv_cgap_filed(k1 + '.' + v1))
# f.write('\t'.join(fieldkeynamelist)+'\n')

def pars_vcfline(vcfline, collist=['CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAT', 'FILTER', 'INFO', 'FORMAT']):
    arr = vcfline.split('\t')
    arr[-1] = arr[-1].strip()

    sid = 0
    v1 = {}
    for i in range(len(arr)):
        try:
            colname = collist[i]
        except IndexError:
            sid += 1
            colname = 'S' + str(sid)

        if colname == "FORMAT":
            sformat = arr[i].split(':')

        if colname == "INFO":
            d = {}
            for f1 in arr[i].split(';'):
                f1 = f1.strip()
                if '=' in f1:
                    arr2 = f1.split('=')
                    d[arr2[0]] = arr2[1]
                else:
                    d[f1] = f1
        if i >= 9:
            d = {}
            arr2 = arr[i].split(':')
            for j in range(len(arr2)):
                d[sformat[j]] = arr2[j]

        if i >= 9 or colname == "INFO":
            v1[colname] = d
        else:
            v1[colname] = arr[i]
    return v1


def pars_info_header(infoheader):
    d = {}
    for s1 in infoheader.strip().split(";"):
        if "=" in s1:
            arr = s1.split('=')
            d[arr[0]] = arr[1].strip().split('|')
    return(d)




import time
from ..util import file_util
from ..util import struct_util
# from ..util.struct_util import get_dict_value as dv
from ..util import vcf_util
from ..util import seq_util
from ..model.datasource import DataSourceList
from ..renderer import VCFRenderer
from ..renderer import JSONRenderer
from .. import _version

INFOIDX = vcf_util.VCF_COL.index('INFO')


class VCFAnnotator():
    vcfheader = None
    datastruct = {}
    opt = {}

    def __init__(self, opt):
        self.vcfheader = None

        self.opt = opt
        self.out = self.opt.out
        self.dslist = DataSourceList(self.opt.ds)
        if self.opt.sourcefile != "":
            self.dslist.sourcefile = self.opt.sourcefile
            self.dslist.sourcefile2 = self.opt.sourcefile
        if self.opt.single_source_mode:
            self.dslist.single_source_mode = self.opt.single_source_mode

        # self.add_hgvs = opt.add_hgvs
        # self.add_variant_class = opt['add_variant_class']
        # self.add_hg19 = opt['add_hg19']
        # self.add_genetable = opt['add_genetable']
        # self.split_multi_allelic_variant = opt['split_multi_allelic_variant']

        self.vcfreader = VCFReader(self.opt)
        self.vcfrenderer = VCFRenderer()
        self.jsonrenderer = JSONRenderer(self.dslist, self.opt.out)


    def get_datafile_header(self, datafile, sourcename, dataformat):
        global DATAHEADER
        header = []
        try:
            header = DATAHEADER[sourcename]
        except KeyError:
            for line in file_util.gzopen(datafile):
                line = line.decode('UTF-8')
                if dataformat == "tab":
                    if line[:len("#CHR")].lower() == "#chr":
                        arr = line[1:].strip().split('\t')
                        header = arr[5:]
                        break
                elif dataformat == "tsv":
                    if line[:len("#CHR")].lower() == "#chr":
                        arr = line[1:].strip().split('\t')
                        header = arr[4:]
                        break
                elif dataformat == "bed":
                    if line[:len("#CHR")].lower() == "#chr":
                        arr = line[1:].strip().split('\t')
                        header = arr
                        break
                elif dataformat == "info":
                    if line[:len("##INFO=")] == "##INFO=":
                        infoID = line[len("##INFO="):].split(',')[0].replace('<ID=', '')
                        target_fieldid = sourcename
                        if '.' in sourcename:
                            target_fieldid = sourcename.split('.')[1].strip()
                        if infoID == target_fieldid:
                            headercont = line.split(
                                'Description="')[-1].split(':')[-1].strip().replace('">', '').replace("'", "")
                            if ' | ' in headercont:
                                header = headercont.split(' | ')
                            else:
                                header = headercont.split('|')
                            break
                elif dataformat == "infovar":
                    if line[:len("##INFO=")] == "##INFO=":
                        infoID = line[len("##INFO="):].split(',')[0].replace('<ID=', '')
                        header.append(infoID)
                if line[0] != '#':
                    break
        DATAHEADER[sourcename] = header
        return header

    def get_annot_header(self):
        cont = ""


        for headertype in ["MUTANNO", "INFO"]:

            if self.opt.add_genoinfo != False:
                cont += vcf_util.get_info_header(headertype, "SAMPLEGENO", "v0.4", "05/07/2020" ,"Sample genotype information", ["NUMGT", "GT", "AD","SAMPLEID"], "samplegeno")
            
            mutanno_fields = []
            if self.opt.add_variant_class:
                mutanno_fields.append("variant_class")
            if self.opt.add_hgvs:
                mutanno_fields.append("hgvsg")
            if self.opt.split_multi_allelic_variant:
                # mutanno_fields.append("samplevariantkey")
                cont += vcf_util.get_info_header(headertype, "MULTIALLELE", _version.VERSION, _version.VERSION_DATE, "sample variant key for multiallelic variant", ["SAMPLEVARIANTKEY"])

            if len(mutanno_fields) > 0:
                cont += vcf_util.get_info_header(headertype, "MUTANNO", _version.VERSION, _version.VERSION_DATE, "mutanno", mutanno_fields)

            if self.opt.add_hg19:
                if self.opt.add_hgvs:
                    cont += vcf_util.get_info_header(headertype, "HG19", "v0.4", "05/07/2020","hg19 coordinates", ["chrom", "pos", "hgvsg"], "hg19")
                else:
                    cont += vcf_util.get_info_header(headertype, "HG19", "v0.4", "05/07/2020", "hg19 coordinates", ["chrom", "pos"], "hg19")
            if self.opt.add_genetable:
                cont += vcf_util.get_info_header(headertype, "GENES","v99","05/07/2020" ,"Gene table", ["ENSG","MOST_SEVERE_TRANSCRIPT","MOST_SEVERE_CONSEQUENCE"], "genes")

        if 'merged_one_field' in self.datastruct.keys() and self.datastruct['merged_one_field'] != '':
            fields = []
            for s1 in self.dslist.source_list:
                if s1.is_available:
                    if s1.fieldselection == "all":
                        if 'use_sourcename_in_fieldname' in self.datastruct.keys():
                            if not self.datastruct['use_sourcename_in_fieldname']:
                                fields.extend(DATAHEADER[s1['name']])
                        else:
                            for f1 in DATAHEADER[s1['name']]:
                                fields.append(s1['name'] + "_" + DATAHEADER[s1['name']][f1])

                    else:
                        for f1 in s1.field_list:
                            if 'use_sourcename_in_fieldname' in self.datastruct.keys() and \
                                    not self.datastruct['use_sourcename_in_fieldname']:
                                if 'name2' in f1.keys():
                                    fields.append(f1['name2'])
                                else:
                                    fields.append(f1['name'])
                            else:
                                if 'name2' in f1.keys():
                                    fields.append(s1['name'] + "_" + f1['name2'])
                                else:
                                    fields.append(s1['name'] + "_" + f1['name'])
            sourcename = self.datastruct['merged_one_field']
            sourcedesc = dv(self.datastruct, 'merged_one_field_desc', '')
            cont += vcf_util.get_info_header("INFO", sourcename, "","", sourcedesc, fields)
        else:
            for s1 in self.dslist.source_list:
                if s1.is_available:
                    fields = []
                    if s1.fieldselection == "all":
                        fields = DATAHEADER[s1.name]
                        fieldlist = []
                        for fieldname in fields:
                            fieldlist.append({'name': fieldname, 'desc': ''})
                        s1['fields'] = fieldlist
                    else:
                        for f1 in s1.field_list:
                            if f1.is_available:
                                fields.append(f1.name2)
                    cont += vcf_util.get_info_header("INFO",s1.name,"","", s1.desc, fields, s1.subembedded)
        return cont

    def get_version_info(self):
        d = {}
        d['MUTANNO'] = {}
        # d['MUTANNO']['MUTANNO'] = {'Version':_version.VERSION, 'Date':_version.VERSION_DATE}
        for s1 in self.dslist.source_list:
            if s1.is_available:
                d['MUTANNO'][s1.name] = {}
                if s1.version != "":
                    d['MUTANNO'][s1.name]['Version'] = s1.version
                if s1.version_date != "":
                    d['MUTANNO'][s1.name]['Date'] = s1.version_date
        return vcf_util.convert_to_metadata(d)

    def open_outpointer(self):
        self.out = file_util.strip_gzext(self.opt.out)
        file_util.check_dir(self.out)
        self.fp = open(self.out, 'w')

    def close_outpointer(self):
        self.fp.close()
        print("Saved " + self.out)


    def write_header(self):
        header = []
        headercont, colheader = self.vcfreader.get_header(keep_only_main_chrom=True)
        header.append(headercont.strip())
        header.append(self.get_version_info().strip())
        header.append(self.get_annot_header().strip())
        header.append(colheader.strip())
        self.fp.write('\n'.join(header) + '\n')

    def run(self):
        self.open_outpointer()
        stime0 = time.time()
        total_varno = 0
        self.write_header()

        ivar = 0
        stime1 = time.time()
        while(not self.vcfreader.eof):
            variant = self.vcfreader.get_variant()
            if self.vcfreader.eof or variant == None:
                break
            variant.set_annot(self.dslist)
            self.fp.write(self.vcfrenderer.render_vcfvariant(variant) + '\n')
            self.jsonrenderer.save_jsonvariant(variant)
            
            ivar += 1
            if ivar % 1000 == 0:
                etime1 = time.time()
                elapsed1 = etime1 - stime1
                log = 'processed... ' + str(ivar) + '/' + str(self.vcfreader.total_variant)
                log += " elapsed:" + str(round(elapsed1, 2)) + "s"
                stime1 = time.time()
                print(log)

        etime0 = time.time()
        elapsed0 = etime0 - stime0
        if ivar > 0:
            print("total:", elapsed0, ", time per variant:", elapsed0 / ivar,
                  ", time for 4M variants:", str(round(elapsed0 / ivar * 4000000 / 3600, 3)) + "hr")
        
        self.close_outpointer()
        
        if self.opt.out.endswith('.gz'):
            file_util.save_gzip(file_util.strip_gzext(self.opt.out))


class VCFReader():
    def __init__(self, opt={}):
        self.vcf = opt.vcf
        self.opt = opt 
        self.add_genoinfo = opt.add_genoinfo
        self.add_hgvs = opt.add_hgvs
        self.add_variant_class = opt.add_variant_class
        
        self.fp = file_util.gzopen(self.vcf)
        self.eof = False
        self.total_variant = 0
        self.i_variant = 0
        self.sampleid_list = []
        self.liftover_hg38_hg19 = None
        if self.opt.add_hg19:
            self.liftover_hg38_hg19 = seq_util.load_liftover(self.opt.chain)

    def is_clean_tag(self, line):
        flag = False
        for prevtag in ["##MUTANNO=<ID=", "##INFO=<ID="]:
            if line[:len(prevtag)] == prevtag:
                for tag in self.opt.clean_tag_list:
                    if tag in line:
                        flag = True
        return flag

    def get_header(self, keep_only_main_chrom=True):
        headercont = ""
        colheader = ""
        for line in file_util.gzopen(self.vcf):
            line = file_util.decodeb(line)
            if line[:2] == "##":
                if line[:len('##contig=')] == '##contig=' and keep_only_main_chrom:
                    contigchrom = line.split(',')[0].replace('##contig=<ID=chr','')
                    if len(contigchrom) < 3:
                        headercont += line
                else:
                    if not self.is_clean_tag(line):
                        headercont += line
            elif line[0] == "#":
                colheader = line
                self.colnames = line[1:].strip().split('\t')
                self.sampleid_list = self.colnames[9:]
            elif line.strip() != '':
                self.total_variant += 1
        return headercont, colheader

    def get_variant(self):
        variant = None
        while True:
            line = self.fp.readline()
            line = file_util.decodeb(line)

            if line.strip() != '':
                if line[0] != '#':
                    variant = VCFVariant(line, self.colnames, self.opt, self.liftover_hg38_hg19)
                    break
            else:
                self.eof = True
                break

        return variant

class VCFVariant:
    def __init__(self, vcfline="", colnames=[], opt={}, liftover_hg38_hg19=None, is_subvariant=False):
        self.opt = opt
        self.vcfline = vcfline
        self.record = vcfline.split('\t')
        self.record[-1] = self.record[-1].strip()
        self.record[INFOIDX] = vcf_util.strip_info(self.record[INFOIDX])
        self.vcfinfo = VCFVariantINFO(self.record[INFOIDX])
        self.chrom = self.record[0]
        self.nchrom = self.record[0].replace('chr','')
        self.pos = int(self.record[1])
        self.ref = self.record[3]
        self.alt = self.record[4]
        self.annotlines = []
        self.colnames = colnames
        self.sampleid_list = colnames[9:]
        self.is_multiallelic = False
        self.is_subvariant = is_subvariant
        self.variant_class = vcf_util.get_variant_class(self.ref, self.alt)
        self.split_variants = []
        self.samplevariantkey = self.chrom + ':' + str(self.pos) + '%20' + self.ref + '/' + self.alt
        self.annotdata = {}
        self.liftover_hg38_hg19 = liftover_hg38_hg19
        self.set_option()
        # print('processing...',self)

    def __str__(self):
        return self.chrom + ':' + str(self.pos)  + '_'  + self.ref + '>' + self.alt

    def set_option(self):
        try:
            if self.opt.add_genoinfo != False and not self.is_subvariant:
                self.add_genotypeinfo_in_infofield(self.opt.add_genoinfo)

            if len(self.opt.clean_tag_list) > 0:
                self.clean_tag_infofield(self.opt.clean_tag_list)

            if self.opt.add_variant_class or self.opt.add_hgvs or self.opt.split_multi_allelic_variant:
                self.add_mutanno_infofield(self.opt.add_variant_class, self.opt.add_hgvs, self.opt.split_multi_allelic_variant)

            if self.opt.add_hg19:
                self.add_hg19_in_infofield(self.liftover_hg38_hg19, self.opt.add_hgvs)
        except AttributeError:
            pass

    def add_genotypeinfo_in_infofield(self, target_sampleid):
        if len(target_sampleid) == 0:
            target_sampleidx = range(9, len(self.colnames))
        else:
            target_sampleidx = [self.sampleid_list.index(sid)+9 for sid in target_sampleid]

        samplegeno = []
        for k in target_sampleidx:
            gtinfo = self.record[k].split(':')
            converted_gtinfo = []
            converted_gtinfo.append(gtinfo[0].replace('|', '/'))
            converted_gtinfo.append(vcf_util.get_genotype(gtinfo[0], self.record[3], self.record[4]))
            converted_gtinfo.append(gtinfo[1].replace(',', '/'))
            converted_gtinfo.append(self.sampleid_list[k-9])
            samplegeno.append('|'.join(converted_gtinfo))
        self.record[INFOIDX] = vcf_util.add_info(self.record[INFOIDX], "SAMPLEGENO=" + ",".join(samplegeno))

    def add_mutanno_infofield(self, add_variant_class=False, add_hgvs=False, split_multi_allelic_variant=False):
        values = []
        if add_variant_class:
            values.append(vcf_util.encode_infovalue(self.variant_class))
        if add_hgvs:
            hgvsg = vcf_util.get_hgvsg(self.chrom, self.pos, self.ref, self.alt)
            values.append(vcf_util.encode_infovalue(hgvsg))
        if split_multi_allelic_variant:
            self.set_split_multiallelic_variant()
            # values.append(self.samplevariantkey)
        if len(values) > 0:
            self.record[INFOIDX] = vcf_util.add_info(self.record[INFOIDX], "MUTANNO=" + '|'.join(values))

        
    def add_hg19_in_infofield(self, liftover_hg38_hg19, add_hgvs=False):
        hg19 = ""
        for hg19coord in seq_util.convert_coordinate(liftover_hg38_hg19, self.chrom, self.pos):
            hg19_chrom = hg19coord[0]
            hg19_pos = hg19coord[1]
            if hg19 != "":
                hg19 += ","
            hg19 += hg19_chrom + '|' + str(hg19_pos)
            if add_hgvs:
                hg19 += '|' + vcf_util.get_hgvsg(hg19_chrom, hg19_pos, self.ref, self.alt)
        if hg19 != "":
            self.record[INFOIDX] = vcf_util.add_info(self.record[INFOIDX], "HG19=" + hg19)
    
    def clean_tag_infofield(self, clean_tag_list):
        rstinfo = ""
        for infofield in self.record[INFOIDX].split(';'):
            tag = infofield.split('=')[0]
            if tag not in clean_tag_list:
                if rstinfo != "":
                    rstinfo += ";"
                rstinfo += infofield
        self.record[INFOIDX] = rstinfo

    def set_split_multiallelic_variant(self):
        if "," in self.alt:
            for record in vcf_util.split_multiallelic_variants(self.record):
                record[INFOIDX] = vcf_util.add_info(record[INFOIDX], "MULTIALLELE=" + vcf_util.encode_infovalue(self.samplevariantkey))
                vcfline = '\t'.join(record)
                subvariant = VCFVariant(vcfline, self.colnames, self.opt, self.liftover_hg38_hg19, is_subvariant=True)

                self.split_variants.append(subvariant)
            self.is_multiallelic = True

    def set_annot(self, dslist):
        self.dslist = dslist
        if self.is_multiallelic:
            for variant in self.split_variants:
                variant.set_annot(dslist)
        else:
            self.annotmerger = dslist.get_variant_annot(self.nchrom, self.pos, ref=self.ref, alt=self.alt, variant=self)
        
    def get_vcf_info(self):
        self.vcfinfo.set_datasourcelist(self.dslist)
        return self.vcfinfo.get_info_dict()



class VCFVariantINFO:
    def __init__(self, recordstring=""):
        self.recordstring = recordstring
        self.data = None
        self.vcfheader = None
        self.dslist = None

    def set_datasourcelist(self, dslist):
        self.dslist = dslist

    def get_info_dict(self):
        if self.data is None:
            if self.dslist is None:
                self.parse_with_only_line()
            elif self.vcfheader is not None:
                self.parse_with_vcfheader()
            elif self.dslist is not None:
                self.parse_with_datasourcelist()
        return self.data

    def parse_with_vcfheader(self):
        pass

    def parse_with_datasourcelist(self):
        pass

    def parse_with_only_line(self):
        d = {}
        for field in self.recordstring.split(';'):
            if '=' in field:
                arr = field.strip().split('=')
                d[arr[0].strip()] = arr[1].split(',')
            else:
                d[field] = True
        self.data = d







from .._options import get_opt_object_from_dict
from ..util import file_util, vcf_util


class TSIReaderSRC:
    def __init__(self, opt={}):
        if isinstance(opt, dict):
            self.opt = get_opt_object_from_dict(opt)
        else:
            self.opt = opt
        self.tsi = self.opt.tsi
        self.tsiheader = []
        self.annotheader = {}
        self.headline = ""
        self.fp = file_util.gzopen(self.tsi)
        self.load_header()

    def load_header(self):
        for line in self.fp:
            line = file_util.decodeb(line)
            if line[0] == "#":
                self.headline = line
                self.tsiheader = line[1:].split('\t')
                self.tsiheader[-1] = self.tsiheader[-1].strip()
                self.annotheader = vcf_util.pars_info_header(self.tsiheader[-1])
                break

    def get_variant(self):
        line = file_util.decodeb(self.fp.readline())
        if line.strip() == "":
            v1 = None
        else:
            v1 = TSIVariant(line, self.tsiheader, self.annotheader)
        return v1


class TSIVariant:
    def __init__(self, line, tsiheader, annotheader):
        record = line.split('\t')
        record[-1] = record[-1].strip()
        self.line = line

        self.chrom = record[0]
        self.pos = int(record[1])
        self.id = record[2].strip()
        self.ref = record[3].strip()
        self.alt = record[4].strip()
        self.qual = record[5]
        self.filter = record[6]
        self.annot = {}
        self.annotheader = annotheader

        self.is_range = False
        self.spos = self.pos
        self.epos = self.pos
        if 'END=' in record[-1]:
            for f1 in record[-1].split(';'):
                if f1[:len('END=')] == 'END=':
                    self.epos = int(f1.replace('END=',''))
                    break
            self.is_range = True
        self.load_annot(record[-1])

    def __str__(self):
        if self.is_range:
            s1 = self.chrom + ':' + str(self.spos) + '-' + str(self.epos)
            # if self.ref != "" and self.alt != "":
            #     s1 += '_' + self.ref + '/' + self.alt
        else:
            s1 = self.chrom + ':' + str(self.pos) + '_' + self.ref + '/' + self.alt
        return s1

    def load_annot(self, cont):
        d = {}
        for sourcefield in cont.split(";"):
            arr = sourcefield.split('=')
            sname = arr[0]
            if sname != 'END':
                try:
                    d[sname]
                except KeyError:
                    d[sname] = []
                for attr in arr[1].split(','):
                    d2 = {}
                    arr2 = attr.split('|')
                    for idx in range(len(self.annotheader[sname])):
                        fieldname = self.annotheader[sname][idx]
                        d2[fieldname] = arr2[idx].strip()
                    d[sname].append(d2)
        self.annot = d
        


import logging

_loggers = {}
_silence_flag = None
_debug_flag = None
_logfile = ""


def get_logger(silence=None, debug=None, logfile=""):
    global _loggers, _silence_flag, _debug_flag, _logfile

    name = "mutanno"
    # print (debug)
    if name not in _loggers:
        _log = logging.getLogger(name)

        if silence is not None:
            _silence_flag = silence
        elif _silence_flag is None:
            _silence_flag = False

        if debug is not None:
            _debug_flag = debug
        elif _debug_flag is not None:
            _debug_flag = False

        if logfile != "":
            _logfile = logfile
        else:
            _logfile = "test.log"

        if _debug_flag:
            _log.setLevel(logging.DEBUG)
            formatter = logging.Formatter('%(asctime)s - %(module)s %(lineno)d: [%(levelname)s] %(message)s')
        else:
            _log.setLevel(logging.INFO)
            formatter = logging.Formatter('%(asctime)s : [%(levelname)s] %(message)s')

        if not _silence_flag:
            sh = logging.StreamHandler()
            sh.setFormatter(formatter)
            _log.addHandler(sh)

        fh = logging.FileHandler(_logfile)
        fh.setFormatter(formatter)
        _log.addHandler(fh)

        _loggers[name] = _log
    return _loggers[name]



#!/usr/bin/env python
# annot.py
# made by Daniel Minseok Kwon
#########################
import tabix
import time
from .util import file_util
from .util import vcf_util
from .util import seq_util
from .util import struct_util
from . import _version

VCFCOLIDX = {'CHROM': 0, 'POS': 1, 'ID': 2, 'REF': 3, 'ALT': 4, 'QUAL': 5, 'FILTER': 6, 'INFO': 7, 'FORMAT': 8}
TSVCOLIDX = {'CHROM': 0, 'POS': 1, 'REF': 2, 'ALT': 3}
DATAHEADER = {}

MAXBUFF = 1000000
INFOCOLIDX = vcf_util.VCF_COL.index('INFO')

# TODO: remove this one


def load_entrez_refseq():
    path = "/home/mk446/bio/mutanno/DATASOURCE/ENSEMBL/hg38/"
    entrezmap = {}
    refseqmap = {}
    try:
        for line in file_util.gzopen(path + 'Homo_sapiens.GRCh38.98.entrez.tsv.gz'):
            line = line.decode('UTF-8')
            arr = line.split('\t')
            arr[-1] = arr[-1].strip()
            entrezmap[arr[0].strip()] = arr[3].strip()

        for line in file_util.gzopen(path + 'Homo_sapiens.GRCh38.98.refseq.sorted.tsv.gz'):
            line = line.decode('UTF-8')
            arr = line.split('\t')
            arr[-1] = arr[-1].strip()
            enst_id = arr[1].strip()
            refseqmap[enst_id] = arr[3].strip()
    except FileNotFoundError:
        pass
    return entrezmap, refseqmap


entrezmap, refseqmap = load_entrez_refseq()


def conv_infovalue(v1):
    if v1 == '.' or v1 == '-':
        v1 = ''
    v1 = v1.replace(";", "%3B").replace("=", "%3D").replace("|", "%7C").replace(",", "%2C")
    v1 = v1.replace('"', "%22").replace('~', "%7E").replace(' ', '%20')
    # v1 = urllib.parse.quote(v1)
    return v1


def get_dict_value(dict1, key1, default):
    rst = default
    if key1 in dict1.keys():
        rst = dict1[key1]
    return rst


def get_sourcename(s1):
    s1name2 = get_dict_value(s1, 'name2', '')
    s1name = get_dict_value(s1, 'name', '')
    if s1name2 != '':
        s1name = s1name2
    return s1name


class AnnotBlock():
    def __init__(self, source, tabixpointer, header, buff=1000):
        self.tabixpointer = tabixpointer
        self.source = source
        self.sourcename = source['name']
        self.header = header
        self.fileformat = source['format']
        self.chrompre = get_dict_value(source, 'chrompre', '')
        self.datatype = get_dict_value(source, 'datatype', '')
        self.buff = buff
        self.block_chrom = ""
        self.block_maxpos = 0
        self.block = {}

    def get_varkey_from_vcfrecord(self, vcfrecord):
        chrom = vcfrecord[0].replace('chr', '').strip()
        pos = int(vcfrecord[1])
        ref = vcfrecord[3].strip()
        alt = vcfrecord[4].strip()
        varkey = vcfrecord[1] + '_' + ref + '>' + alt
        return chrom, pos, ref, alt, varkey

    def get_annotmap(self, vcfrecord, varkeyset):
        chrom, pos, ref, alt, varkey = self.get_varkey_from_vcfrecord(vcfrecord)

        if chrom != self.block_chrom or pos > self.block_maxpos:
            print("load block", str(pos) + '~' + str(pos + MAXBUFF), self.sourcename, self.block_maxpos)
            self.load_annotblock(chrom, pos, varkeyset)

        try:
            annotmap = self.block[varkey]
        except KeyError:
            annotmap = {}
        return annotmap

    def load_annotblock(self, chrom, pos, varkeyset={}):
        self.block = {}
        if self.fileformat == "tab" or self.fileformat == "tsv" or self.fileformat == "bed":
            self.load_annotblock_tabformat(chrom, pos, varkeyset)
            # annotmap = self.load_annotdata_tabformat(chrom, pos, ref, alt)
        elif self.fileformat == "info":
            self.load_annotblock_infoformat(chrom, pos, varkeyset)
            # annotmap = self.load_annotdata_infoformat(chrom, pos, ref, alt)
        elif self.fileformat == "infovar":
            self.load_annotblock_infovarformat(chrom, pos, varkeyset)
            # annotmap = self.load_annotdata_infovarformat(chrom, pos, ref, alt)

    def load_annotblock_tabformat(self, chrom, pos, varkeyset={}):
        if self.fileformat == 'tsv':
            colidx = TSVCOLIDX
            sidx = 4
        elif self.fileformat == 'tab':
            colidx = VCFCOLIDX
            sidx = 5
        elif self.fileformat == 'bed':
            sidx = 0

        recs = self.tabixpointer.query(self.chrompre + chrom, pos, pos + MAXBUFF)
        for r1 in recs:
            d = {}
            for k in range(len(self.header)):
                d[self.header[k]] = conv_infovalue(r1[k + sidx].strip())
            varkey = r1[colidx['POS']] + '_' + r1[colidx['REF']] + '>' + r1[colidx['ALT']]

            if len(varkeyset.keys()) > 0:
                try:
                    varkeyset[varkey]
                    try:
                        self.block[varkey].append(d)
                    except KeyError:
                        self.block[varkey] = [d]
                except KeyError:
                    pass
            else:
                try:
                    self.block[varkey].append(d)
                except KeyError:
                    self.block[varkey] = [d]

            # print(list(varkeyset.keys())[:3])
            # print(pos, varkey, len(self.block.keys()))
            if len(self.block.keys()) % 10 == 0:
                print(len(self.block.keys()), varkey)
            if len(self.block.keys()) >= self.buff:
                self.block_maxpos = int(r1[colidx['POS']])
                self.block_chrom = chrom
                break

    def load_annotblock_infovarformat(self, chrom, pos, varkeyset={}):
        try:
            # recs = self.tabixpointer.query(self.chrompre+chrom, pos, pos+self.buff)
            recs = self.tabixpointer.query(self.chrompre + chrom, pos, pos + MAXBUFF)
            for r1 in recs:
                d = {}
                for infofield in r1[VCFCOLIDX['INFO']].split(';'):
                    if '=' in infofield:
                        infoarr = infofield.split('=')
                        fieldid = infoarr[0]
                        fieldvalue = infoarr[1]
                        d[fieldid] = conv_infovalue(fieldvalue.strip())

                varkey = r1[VCFCOLIDX['POS']] + '_' + r1[VCFCOLIDX['REF']] + '>' + r1[VCFCOLIDX['ALT']]
                if len(varkeyset.keys()) > 0:
                    try:
                        varkeyset[varkey]
                        try:
                            self.block[varkey].append(d)
                        except KeyError:
                            self.block[varkey] = [d]
                    except KeyError:
                        pass
                else:
                    try:
                        self.block[varkey].append(d)
                    except KeyError:
                        self.block[varkey] = [d]

                if len(self.block.keys()) % 10 == 0:
                    print(len(self.block.keys()), varkey)

                if len(self.block.keys()) >= self.buff:
                    self.block_maxpos = int(r1[VCFCOLIDX['POS']])
                    self.block_chrom = chrom
                    break
        except tabix.TabixError:
            pass

    def load_annotdata_tabformat(self, chrom, pos, ref, alt):
        recs = self.tabixpointer.query(self.chrompre + chrom, pos, pos)
        sdata = []
        for r1 in recs:
            if self.fileformat == 'tsv':
                colidx = TSVCOLIDX
                sidx = 4
            elif self.fileformat == 'tab':
                colidx = VCFCOLIDX
                sidx = 5
            elif self.fileformat == 'bed':
                sidx = 0

            if self.datatype == "region" or r1[colidx['REF']] == ref and r1[colidx['ALT']] == alt:
                d = {}
                for k in range(len(self.header)):
                    d[self.header[k]] = conv_infovalue(r1[k + sidx].strip())
                sdata.append(d)
        return sdata

    def load_annotdata_infoformat(self, chrom, pos, ref, alt):
        recs = self.tabixpointer.query(self.chrompre + chrom, pos, pos)
        sdata = []

        target_fieldid = self.sourcename
        if '.' in self.sourcename:
            target_fieldid = self.sourcename.split('.')[1].strip()

        for r1 in recs:
            if r1[VCFCOLIDX['REF']] == self.ref and r1[VCFCOLIDX['ALT']] == self.alt:
                for infofield in r1[VCFCOLIDX['INFO']].split(';'):
                    infoarr = infofield.split('=')
                    fieldid = infoarr[0]
                    if target_fieldid == fieldid:
                        for subfield in infoarr[1].strip().split(','):
                            arr = subfield.split('|')
                            if self.sourcename == "SNPEFF.LOF" or self.sourcename == "SNPEFF.NMD":
                                arr[0] = arr[0][1:]         # remove '('
                                arr[-1] = arr[-1][:-1]      # remove ')'
                            d = {}
                            for k in range(len(self.header)):
                                d[self.header[k]] = conv_infovalue(arr[k].strip())
                            sdata.append(d)
        return sdata

    def load_annotdata_infovarformat(self, chrom, pos, ref, alt):
        sdata = []
        try:
            recs = self.tabixpointer.query(self.chrompre + chrom, pos, pos)
            # recs = self.tabixpointer.query(self.chrompre+chrom, pos, pos+10*1000*1000)
            for r1 in recs:
                if r1[VCFCOLIDX['REF']] == ref and r1[VCFCOLIDX['ALT']] == alt:
                    d = {}
                    for infofield in r1[VCFCOLIDX['INFO']].split(';'):
                        if '=' in infofield:
                            infoarr = infofield.split('=')
                            fieldid = infoarr[0]
                            fieldvalue = infoarr[1]
                            d[fieldid] = conv_infovalue(fieldvalue.strip())
                    sdata.append(d)
        except tabix.TabixError:
            pass
        return sdata


class AnnotMapBlock():
    subembedded_keymap = {}

    def __init__(self, datastruct, datafileinfo, buff):
        self.datafileinfo = datafileinfo
        self.datastruct = datastruct
        self.subembedded_keymap = {}
        self.set_annotblock(buff)

    def set_annotblock(self, buff):
        self.annotblock = {}
        ds = self.datastruct
        for s1 in ds['source']:
            tp = self.datafileinfo['tps'][s1['name']]
            header = self.datafileinfo['headers'][s1['name']]
            self.annotblock[s1['name']] = AnnotBlock(s1, tp, header, buff)

    def get_annotmap(self, vcfrecord, varkeyset={}):
        annotmap = {}
        ds = self.datastruct
        for s1 in ds['source']:
            annotmap[s1['name']] = self.annotblock[s1['name']].get_annotmap(vcfrecord, varkeyset)
        return annotmap

    def set_variantsample_annot(self):
        pass

    def set_transcript_annot(self):
        pass

    def set_gene_annot(self):
        pass

    def get_annotvcf_block(self, vcfblock):
        vcfblockcont = ""
        varkeyset = {}
        for i in range(len(vcfblock)):
            v1 = vcfblock[i]
            varkeyset[v1[1] + '_' + v1[3] + '>' + v1[4]] = 1

        for i in range(len(vcfblock)):
            vcfrecord = self.get_annotvcf_record(vcfblock[i], varkeyset)
            vcfblockcont += vcfrecord + '\n'
        return vcfblockcont

    def get_annotvcf_record(self, vcfrecord, varkeyset={}):
        vcfrecord = self.cleanup_infofield(vcfrecord)
        annotmap = self.get_annotmap(vcfrecord, varkeyset)
        if 'merged_one_field' in self.datastruct.keys() and self.datastruct['merged_one_field'] != '':
            vcfrecord = self.update_info_with_annotmap_in_one_field(vcfrecord, annotmap)
        else:
            vcfrecord = self.update_info_with_annotmap(vcfrecord, annotmap)
        return '\t'.join(vcfrecord)

    def cleanup_infofield(self, vcfrecord):
        inforec = vcfrecord[VCFCOLIDX['INFO']]
        if inforec != '' and inforec != '.' and inforec[-1] != ';':
            inforec += ';'
        if inforec == '.':
            inforec = ''
        vcfrecord[VCFCOLIDX['INFO']] = inforec
        return vcfrecord

    def update_info_with_annotmap_in_one_field(self, vcfrecord, annotmap):
        merged_annot = {}
        multi_fields_sourcename = ""
        no_multi_fields_source = 0
        for s1 in self.datastruct['source']:
            fieldselection = get_dict_value(s1, 'fieldselection', '')
            annotlist = []
            if s1['name'] in annotmap.keys():
                for subfield in annotmap[s1['name']]:
                    subannotlist = []
                    if fieldselection == "all":
                        for f1 in subfield.keys():
                            subannotlist.append(subfield[f1].strip())
                    else:
                        for finfo in s1['fields']:
                            f1 = finfo['name']
                            try:
                                a1 = subfield[f1].strip()
                            except KeyError:
                                a1 = ''
                            subannotlist.append(a1)
                    annotlist.append('|'.join(subannotlist))
            merged_annot[s1['name']] = annotlist
            if multi_fields_sourcename == "":
                multi_fields_sourcename = s1['name']
            if len(annotlist) > 1:
                no_multi_fields_source += 1
                multi_fields_sourcename = s1['name']

        if no_multi_fields_source <= 1:
            merged_annotlist = []
            for annotfield in merged_annot[multi_fields_sourcename]:
                for s1name in merged_annot.keys():
                    if s1name != multi_fields_sourcename:
                        if len(merged_annot[s1name]) > 0:
                            annotfield += '|' + merged_annot[s1name][0]
                        else:
                            annotfield += '|'
                merged_annotlist.append(annotfield)
            vcfrecord[VCFCOLIDX['INFO']] += self.datastruct['merged_one_field'] + '=' + ','.join(merged_annotlist) + ';'
        return vcfrecord

    def update_info_with_annotmap(self, vcfrecord, annotmap):
        for s1 in self.datastruct['source']:
            flag_post = False
            fieldselection = get_dict_value(s1, 'fieldselection', '')
            subembedded = get_dict_value(s1, 'subembedded', '')
            annotlist = []
            annotlistmap = {}

            flag_make_subembedded_keymap = False
            if subembedded != '':
                try:
                    self.subembedded_keymap[subembedded]
                except KeyError:
                    self.subembedded_keymap[subembedded] = []
                    flag_make_subembedded_keymap = True

            if s1['name'] == "VEP":
                annotlist = self.get_vep_annotlist(fieldselection, s1, annotmap)
                flag_post = True
                # print("subembedded_keymap:", self.subembedded_keymap)
            elif s1['name'] in annotmap.keys():
                annotlist = []
                for subfield in annotmap[s1['name']]:
                    subannotlist = []
                    if fieldselection == "all":
                        for f1 in subfield.keys():
                            subannotlist.append(subfield[f1].strip())
                    else:
                        for finfo in s1['fields']:
                            if get_dict_value(finfo, 'is_available', True):
                                f1 = finfo['name']
                                try:
                                    a1 = subfield[f1].strip()
                                except KeyError:
                                    a1 = ''

                                if get_dict_value(finfo, 'is_list', False):
                                    a1 = "~".join(a1.split(conv_infovalue(finfo['delimiter'])))
                                subannotlist.append(a1)

                                if get_dict_value(finfo, 'subembedded_key', False):
                                    if flag_make_subembedded_keymap:
                                        self.subembedded_keymap[subembedded].append(a1)
                                    else:
                                        subembedded_key = a1
                                        print('self.subembedded_keymap_index:',
                                              self.subembedded_keymap[subembedded].index(a1))

                                if a1 != '':
                                    flag_post = True
                    if len(subannotlist) > 0:
                        annotlist.append('|'.join(subannotlist))
                        if subembedded != '' and not flag_make_subembedded_keymap:
                            annotlistmap[self.subembedded_keymap[subembedded].index(
                                subembedded_key)] = '|'.join(subannotlist)
                            subannotlist_len = len(subannotlist)

            # TODO: ##########################################
            # print (">>>>>>>>>>>>>>",s1['name'],s1['fields'],annotlist)
            if flag_post:
                sourcename = get_sourcename(s1)
                if subembedded != '' and not flag_make_subembedded_keymap:
                    annotlist = self.get_annotlist_from_annotlistmap_in_subembedded(
                        annotlistmap, subannotlist_len, subembedded)
                vcfrecord[VCFCOLIDX['INFO']] += sourcename + '=' + ','.join(annotlist) + ';'
            return vcfrecord

    def get_annotlist_from_annotlistmap_in_subembedded(self, annotlistmap, subannotlist_len, subembedded):
        annotlist = []

        for k in range(len(self.subembedded_keymap[subembedded])):
            try:
                v1 = annotlistmap[k]
            except KeyError:
                v1 = ""
                for j in range(subannotlist_len):
                    v1 += '|'
            annotlist.append(v1)
        return annotlist

    def get_vep_annotlist(self, fieldselection, s1, annotmap):
        global entrezmap, refseqmap
        subembedded = get_dict_value(s1, 'subembedded', '')
        annotlist = []
        if s1['name'] in annotmap.keys():
            for subfield in annotmap[s1['name']]:
                subannotlist = []
                if fieldselection == "all":
                    for f1 in subfield.keys():
                        subannotlist.append(subfield[f1].strip())
                else:
                    if subfield['Feature'][:len('ENST')] == "ENST":
                        try:
                            entrez_xref = entrezmap[subfield['Gene']]
                        except KeyError:
                            entrez_xref = ''
                        try:
                            refseq_xref = refseqmap[subfield['Feature']]
                        except KeyError:
                            refseq_xref = ''

                        subfield['Feature_ncbi'] = refseq_xref
                        subfield['Gene_ncbi'] = entrez_xref

                        for finfo in s1['fields']:
                            if get_dict_value(finfo, 'is_available', True):
                                f1 = finfo['name']
                                try:
                                    a1 = subfield[f1].strip()
                                except KeyError:
                                    a1 = ''

                                if get_dict_value(finfo, 'subembedded_key', False):
                                    self.subembedded_keymap[subembedded].append(a1)

                                if get_dict_value(finfo, 'is_list', False):
                                    a1 = "~".join(a1.split(conv_infovalue(finfo['delimiter'])))
                                subannotlist.append(a1)

                if len(subannotlist) > 0:
                    annotlist.append('|'.join(subannotlist))
        return annotlist


class AnnotMap():
    def __init__(self, datastruct, datafileinfo):
        self.datastruct = datastruct
        self.tabixpointer = None
        self.gzpointer = None
        self.tchrom = ""
        self.flag_gzsource = True
        self.prev_record = []
        self.defaultvalue = ""
        self.load_defaultvalue()

    def load_tabixpointer(self):
        print(self.datastruct['sourcefile'].replace("#CHROM#", self.tchrom))
        self.tabixpointer = tabix.open(self.datastruct['sourcefile'].replace("#CHROM#", self.tchrom))

    def load_gzpointer(self):
        if self.datastruct['sourcefile'].endswith('.gz'):
            self.flag_gzsource = True
            self.gzpointer = file_util.gzopen(self.datastruct['sourcefile'].replace("#CHROM#", self.tchrom))
        else:
            self.flag_gzsource = False
            self.gzpointer = open(self.datastruct['sourcefile'].replace("#CHROM#", self.tchrom), 'r')

    def load_defaultvalue(self):
        for source in self.datastruct['source']:
            flag_add = False
            defaultvaluelist = []
            for field in source['fields']:
                defaultvalue = get_dict_value(field, 'default', "")
                defaultvaluelist.append(defaultvalue)
                if defaultvalue != "":
                    flag_add = True
            if flag_add:
                if self.defaultvalue != "" and self.defaultvalue[-1] != ';':
                    self.defaultvalue += ";"
                self.defaultvalue += source['name'] + '=' + "|".join(defaultvaluelist)

    def get_annotrecord_from_gzpoint(self):
        rec = []
        if len(self.prev_record) > 2:
            rec = self.prev_record
        else:
            while True:
                line = self.gzpointer.readline()
                if self.flag_gzsource:
                    line = line.decode('UTF-8')
                if line[0] != '#':
                    rec = line.split('\t')
                    rec[-1] = rec[-1].strip()
                    break
                if line.strip() == '':
                    break
        return rec

    def write_annotvcf_with_vcfblock(self, vcfblock, fp):
        stime = time.time()
        for b1 in vcfblock:
            chrom = b1[0].replace('chr', '')
            pos = int(b1[1])
            ref = b1[3]
            alt = b1[4]
            if chrom != self.tchrom:
                self.tchrom = chrom
                self.load_gzpointer()

            if b1[7] == '.':
                b1[7] = ""

            while True:
                r1 = self.get_annotrecord_from_gzpoint()
                apos = int(r1[1])
                if apos == pos:
                    if r1[3] == ref and r1[4] == alt:
                        if b1[7].strip() != '':
                            b1[7] += ';'
                        b1[7] += r1[5]
                        self.prev_record = []
                        break
                elif apos > pos:
                    break
                else:
                    self.prev_record = []
            fp.write('\t'.join(b1) + '\n')
        etime = time.time()
        elapsed = etime - stime
        print('processed...', len(vcfblock), "elapsed:", elapsed)
        return len(vcfblock)

    # slow but useful for sparse VCF
    def write_annotvcf_with_vcfblock_tabix(self, vcfblock, fp, is_rm_unannotated=False):
        for b1 in vcfblock:
            chrom = b1[0].replace('chr', '')
            pos = int(b1[1])
            ref = b1[3]
            alt = b1[4]
            if chrom != self.tchrom and ('#CHROM#' in self.datastruct['sourcefile'] or self.tabixpointer is None):
                self.tchrom = chrom
                self.load_tabixpointer()

            flag_add = False
            # print(chrom, b1)
            try:
                recs = self.tabixpointer.query(chrom, pos - 1, pos + 1)
            except tabix.TabixError:
                recs = []
            for r1 in recs:
                if int(r1[1]) == pos and r1[3] == ref and r1[4] == alt:
                    if b1[INFOCOLIDX] == '.':
                        b1[INFOCOLIDX] = ""
                    if b1[INFOCOLIDX].strip() != '' and b1[INFOCOLIDX][-1] != ';':
                        b1[INFOCOLIDX] += ';'
                    b1[INFOCOLIDX] += r1[5]
                    flag_add = True
                    break

            if flag_add:
                fp.write('\t'.join(b1) + '\n')
            elif not is_rm_unannotated:
                if b1[INFOCOLIDX] != '' and b1[INFOCOLIDX][-1] != ';':
                    b1[INFOCOLIDX] += ';'
                b1[INFOCOLIDX] += self.defaultvalue
                fp.write('\t'.join(b1) + '\n')

        return len(vcfblock)


class VCFBlockReader():
    def __init__(self, opt={}):
        self.vcf = struct_util.get_dict_value(opt, 'vcf', '')
        self.blocksize = struct_util.get_dict_value(opt, 'blocksize', 1000)
        self.add_genoinfo = struct_util.get_dict_value(opt, 'add_genoinfo', False)
        self.add_hgvs = struct_util.get_dict_value(opt, 'add_hgvs', False)
        self.add_genetable = struct_util.get_dict_value(opt, 'add_genetable', False)
        self.clean_tag_list = struct_util.get_dict_value(opt, 'clean_tag_list', [])
        self.add_hg19 = struct_util.get_dict_value(opt, 'add_hg19', False)
        if self.add_hg19:
            self.liftover_hg38_hg19 = seq_util.load_liftover(struct_util.get_dict_value(opt, 'chain', ''))
        self.fp = file_util.gzopen(self.vcf)
        self.eof = False
        self.total_variant = 0
        self.i_variant = 0
        self.sampleid_list = []

    def is_clean_tag(self, line):
        flag = False
        for prevtag in ["##MUTANNO=<ID=", "##INFO=<ID="]:
            if line[:len(prevtag)] == prevtag:
                for tag in self.clean_tag_list:
                    if tag in line:
                        flag = True
        return flag

    def get_header(self, inserted_header):
        headercont = ""
        for line in file_util.gzopen(self.vcf):
            line = file_util.decodeb(line)
            if line[0] == "#":
                if line[:len('#CHROM')] == "#CHROM":
                    headercont += inserted_header
                    self.sampleid_list = line.strip().split('\t')[9:]
                if line[:len('##contig=')] == '##contig=':
                    contigchrom = line.split(',')[0].replace('##contig=<ID=chr', '')
                    if len(contigchrom) < 3:
                        headercont += line
                else:
                    if not self.is_clean_tag(line):
                        headercont += line
            elif line.strip() != '':
                self.total_variant += 1
        return headercont

    def add_genotypeinfo_in_infofield(self, vcfrecord, target_sampleid):

        if len(target_sampleid) == 0:
            target_sampleidx = range(9, len(vcfrecord))
        else:
            target_sampleidx = [self.sampleid_list.index(sid)+9 for sid in target_sampleid]

        samplegeno = []
        for k in target_sampleidx:
            gtinfo = vcfrecord[k].split(':')
            converted_gtinfo = []
            converted_gtinfo.append(gtinfo[0].replace('|', '/'))
            converted_gtinfo.append(vcf_util.get_genotype(gtinfo[0], vcfrecord[3], vcfrecord[4]))
            converted_gtinfo.append(gtinfo[1].replace(',', '/'))
            converted_gtinfo.append(self.sampleid_list[k-9])
            samplegeno.append('|'.join(converted_gtinfo))
        return self.add_infofield(vcfrecord, "SAMPLEGENO", ",".join(samplegeno))

    def get_hgvsg(self, chrom, pos, ref, alt):
        try:
            hgvsg = vcf_util.CHROM_HGVS[chrom.replace('chr', '')] + ':g.'
            if len(ref) > len(alt):
                if len(ref) == 2:
                    hgvsg += str(pos + 1) + 'del'
                else:
                    hgvsg += str(pos + 1) + '_' + str(pos + len(ref) - 1) + 'del'
            elif len(ref) < len(alt):
                hgvsg += str(pos) + '_' + str(pos + 1) + 'ins' + alt[1:]
            else:
                hgvsg += str(pos) + ref + '>' + alt
        except KeyError:
            hgvsg = ""
        return hgvsg

    def add_hgvs_in_infofield(self, vcfrecord):
        chrom = vcfrecord[0]
        pos = int(vcfrecord[1])
        ref = vcfrecord[3]
        alt = vcfrecord[4]

        hgvsg = self.get_hgvsg(chrom, pos, ref, alt)

        return self.add_infofield(vcfrecord, "HGVS", hgvsg)

    def add_hg19_in_infofield(self, vcfrecord):
        chrom = vcfrecord[0]
        pos = int(vcfrecord[1])
        ref = vcfrecord[3]
        alt = vcfrecord[4]
        hg19 = ""
        for hg19coord in seq_util.convert_coordinate(self.liftover_hg38_hg19, chrom, pos):
            hg19_chrom = hg19coord[0]
            hg19_pos = hg19coord[1]

            if hg19 != "":
                hg19 += ","

            hg19 += hg19_chrom + '|' + str(hg19_pos)
            if self.add_hgvs:
                hg19 += '|' + self.get_hgvsg(hg19_chrom, hg19_pos, ref, alt)

        return self.add_infofield(vcfrecord, "HG19", hg19)

    def add_infofield(self, vcfrecord, fieldname, fieldcont):
        infofield = vcfrecord[INFOCOLIDX]
        if infofield == ".":
            infofield = ""
        if infofield != "" and infofield[-1] != ";":
            infofield += ";"

        if fieldcont != "":
            infofield += fieldname + "=" + fieldcont
        vcfrecord[INFOCOLIDX] = infofield
        return vcfrecord

    def add_genetable_in_infofield(self, vcfrecord):
        # chrom = vcfrecord[0]
        # pos = int(vcfrecord[1])
        # ref = vcfrecord[3]
        # alt = vcfrecord[4]
        # hg19 = ""
        # for hg19coord in seq_util.convert_coordinate(chrom, pos):
        #     hg19_chrom = hg19coord[0]
        #     hg19_pos = hg19coord[1]

        #     if hg19 != "":
        #         hg19 += ","

        #     hg19 += hg19_chrom + '|' + str(hg19_pos)
        #     if self.add_hgvs:
        #         hg19 += '|' + self.get_hgvsg(hg19_chrom, hg19_pos, ref, alt)

        # if vcfrecord[7] == ".":
        #     vcfrecord[7] = ""
        # if vcfrecord[7] != "" and vcfrecord[7][-1] != ";":
        #     vcfrecord[7] += ";"

        # if hg19 != "":
        #     vcfrecord[7] += "HG19=" + hg19
        return vcfrecord

    def clean_tag_infofield(self, vcfrecord):
        rstinfo = ""
        for infofield in vcfrecord[7].split(';'):
            tag = infofield.split('=')[0]
            if tag not in self.clean_tag_list:
                if rstinfo != "":
                    rstinfo += ";"
                rstinfo += infofield
        vcfrecord[7] = rstinfo
        return vcfrecord

    def get_block(self):
        block = []
        while True:
            line = self.fp.readline()
            line = file_util.decodeb(line)

            if line.strip() != '':
                if line[0] != '#':
                    vcfrecord = line.split('\t')
                    vcfrecord[-1] = vcfrecord[-1].strip()
                    if type(self.add_genoinfo) == list:
                        vcfrecord = self.add_genotypeinfo_in_infofield(vcfrecord, self.add_genoinfo)
                    if self.add_hgvs:
                        vcfrecord = self.add_hgvs_in_infofield(vcfrecord)
                    if self.add_hg19:
                        vcfrecord = self.add_hg19_in_infofield(vcfrecord)
                    if self.add_genetable:
                        vcfrecord = self.add_genetable_in_infofield(vcfrecord)
                    if len(self.clean_tag_list) > 0:
                        vcfrecord = self.clean_tag_infofield(vcfrecord)
                    if "," in vcfrecord[4]:
                        for a1 in vcf_util.split_multiallelic_variants(vcfrecord):
                            block.append(a1)
                            self.i_variant += 1
                    else:
                        block.append(vcfrecord)
                        self.i_variant += 1
            else:
                self.eof = True
                break
            if len(block) >= self.blocksize:
                break

        return block


class AnnotVCF():
    vcfheader = None
    datafileinfo = {}
    datastruct = {}
    opt = {}

    def __init__(self, opt):
        self.vcfheader = None
        self.datafileinfo = {}
        self.datafileinfo['files'] = {}
        self.datafileinfo['tps'] = {}
        self.datafileinfo['headers'] = {}

        self.opt = opt
        self.datastruct = file_util.load_json(opt['ds'])
        if self.opt['sourcefile'] != '':
            self.datastruct['sourcefile'] = self.opt['sourcefile']
        self.blocksize = opt['blocksize']
        self.add_genoinfo = opt['add_genoinfo']
        self.add_hgvs = opt['add_hgvs']
        self.add_hg19 = opt['add_hg19']
        self.add_genetable = opt['add_genetable']
        self.split_multi_allelic_variant = opt['split_multi_allelic_variant']

        self.set_datafile()

    def get_datafile_header(self, datafile, sourcename, dataformat):
        global DATAHEADER
        header = []
        try:
            header = DATAHEADER[sourcename]
        except KeyError:
            for line in file_util.gzopen(datafile):
                line = line.decode('UTF-8')
                if dataformat == "tab":
                    if line[:len("#CHR")].lower() == "#chr":
                        arr = line[1:].strip().split('\t')
                        header = arr[5:]
                        break
                elif dataformat == "tsv":
                    if line[:len("#CHR")].lower() == "#chr":
                        arr = line[1:].strip().split('\t')
                        header = arr[4:]
                        break
                elif dataformat == "bed":
                    if line[:len("#CHR")].lower() == "#chr":
                        arr = line[1:].strip().split('\t')
                        header = arr
                        break
                elif dataformat == "info":
                    if line[:len("##INFO=")] == "##INFO=":
                        infoID = line[len("##INFO="):].split(',')[0].replace('<ID=', '')
                        target_fieldid = sourcename
                        if '.' in sourcename:
                            target_fieldid = sourcename.split('.')[1].strip()
                        if infoID == target_fieldid:
                            headercont = line.split(
                                'Description="')[-1].split(':')[-1].strip().replace('">', '').replace("'", "")
                            if ' | ' in headercont:
                                header = headercont.split(' | ')
                            else:
                                header = headercont.split('|')
                            break
                elif dataformat == "infovar":
                    if line[:len("##INFO=")] == "##INFO=":
                        infoID = line[len("##INFO="):].split(',')[0].replace('<ID=', '')
                        header.append(infoID)
                if line[0] != '#':
                    break
        DATAHEADER[sourcename] = header
        return header

    def set_datafile(self, chrom='1'):
        for s1 in self.datastruct['source']:
            if 'datafile' in s1.keys():
                datafile = s1['datafile'].replace('#CHROM#', chrom)
                if file_util.is_exist(datafile):
                    tp = tabix.open(datafile)
                    header = self.get_datafile_header(datafile, s1['name'], s1['format'])
                    self.datafileinfo['files'][s1['name']] = datafile
                    self.datafileinfo['tps'][s1['name']] = tp
                    self.datafileinfo['headers'][s1['name']] = header
                else:
                    # print('Error: File not exist.', datafile)
                    pass

    def get_annot_header(self):
        cont = ""

        for headertype in ["MUTANNO", "INFO"]:
            if self.add_genoinfo:
                cont += vcf_util.get_info_header(headertype, "SAMPLEGENO", "v0.4", "05/07/2020",
                                                 "Sample genotype information", ["NUMGT", "GT", "AD", "SAMPLEID"],
                                                 "samplegeno")
            if self.add_hgvs:
                cont += vcf_util.get_info_header(headertype, "HGVS", "v0.4", "05/07/2020", "HGVSG", ["hgvsg"])
            if self.add_hg19:
                if self.add_hgvs:
                    cont += vcf_util.get_info_header(headertype, "HG19", "v0.4", "05/07/2020",
                                                     "hg19 coordinates", ["chrom", "pos", "hgvsg"], "hg19")
                else:
                    cont += vcf_util.get_info_header(headertype, "HG19", "v0.4",
                                                     "05/07/2020", "hg19 coordinates", ["chrom", "pos"], "hg19")
            if self.add_genetable:
                cont += vcf_util.get_info_header(headertype, "GENES", "v99", "05/07/2020", "Gene table",
                                                 ["ENSG", "MOST_SEVERE_TRANSCRIPT", "MOST_SEVERE_CONSEQUENCE"], "genes")

            if self.split_multi_allelic_variant:
                cont += vcf_util.get_info_header(headertype, "multiallele", "v0.4", "05/07/2020",
                                                 "sample-variant key for multi-allelic variants", ["SAMPLEVARIANTKEY"])

        if 'merged_one_field' in self.datastruct.keys() and self.datastruct['merged_one_field'] != '':
            fields = []
            for s1 in self.datastruct['source']:
                if get_dict_value(s1, 'is_available', True):
                    fieldselection = get_dict_value(s1, 'fieldselection', '')
                    if fieldselection == "all":
                        if 'use_sourcename_in_fieldname' in self.datastruct.keys():
                            if not self.datastruct['use_sourcename_in_fieldname']:
                                fields.extend(DATAHEADER[s1['name']])
                        else:
                            for f1 in DATAHEADER[s1['name']]:
                                fields.append(s1['name'] + "_" + DATAHEADER[s1['name']][f1])

                    else:
                        for f1 in s1['fields']:
                            if 'use_sourcename_in_fieldname' in self.datastruct.keys() and \
                                    not self.datastruct['use_sourcename_in_fieldname']:
                                if 'name2' in f1.keys():
                                    fields.append(f1['name2'])
                                else:
                                    fields.append(f1['name'])
                            else:
                                if 'name2' in f1.keys():
                                    fields.append(s1['name'] + "_" + f1['name2'])
                                else:
                                    fields.append(s1['name'] + "_" + f1['name'])
            sourcename = self.datastruct['merged_one_field']
            sourcedesc = get_dict_value(self.datastruct, 'merged_one_field_desc', '')
            cont += vcf_util.get_info_header("INFO", sourcename, "", "", sourcedesc, fields)
        else:
            for s1 in self.datastruct['source']:
                if get_dict_value(s1, 'is_available', True):
                    fieldselection = get_dict_value(s1, 'fieldselection', '')
                    fields = []
                    if fieldselection == "all":
                        fields = DATAHEADER[s1['name']]
                        fieldlist = []
                        for fieldname in fields:
                            fieldlist.append({'name': fieldname, 'desc': ''})
                        s1['fields'] = fieldlist
                    else:
                        for f1 in s1['fields']:
                            if get_dict_value(f1, 'is_available', True):
                                if get_dict_value(f1, 'name2', False):
                                    fields.append(f1['name2'])
                                else:
                                    fields.append(f1['name'])
                    sourcename = get_sourcename(s1)
                    sourcedesc = get_dict_value(s1, 'desc', '')
                    sourcesubembed = get_dict_value(s1, 'subembedded', '')
                    cont += vcf_util.get_info_header("INFO", sourcename, "", "", sourcedesc, fields, sourcesubembed)
        return cont

    def get_version_info(self):
        ds = self.datastruct

        cont = ""
        cont += '##MUTANNO=<ID=MUTANNO'
        cont += ',Version="' + _version.VERSION + '"'
        cont += ',Date="' + _version.VERSION_DATE + '"'

        if 'data_version' in ds.keys():
            cont += ',DataVersion="' + ds['data_version'] + '"'
        if 'data_version_date' in ds.keys():
            cont += ',DataDate="' + ds['data_version_date'] + '"'

        cont += '>\n'

        for s1 in ds['source']:
            sourcename = get_sourcename(s1)
            cont += '##MUTANNO=<ID=' + sourcename
            if 'version' in s1.keys():
                cont += ',Version="' + s1['version'] + '"'
            if 'version_date' in s1.keys():
                cont += ',Date="' + s1['version_date'] + '"'
            cont += '>\n'
        return cont

    def run_by_loading_source(self):
        stime = time.time()
        annotmap = AnnotMapBlock(self.datastruct, self.datafileinfo, self.blocksize)
        etime = time.time()
        elapsed = etime - stime
        print(cnt, elapsed, 'sec')

        '''
            varno += 1
            # if self.vcfheader is None:
            #     self.vcfheader = vcf_util.VCFHEADER(headercont)
            #     f.write(self.vcfheader.header.strip() + '\n')
            arr = line.split('\t')
            arr[-1] = arr[-1].strip()

            chrom = arr[0].replace('chr', '')
            if prev_chrom != chrom:
                self.set_datafile(chrom)

            if flag_block:
                vcfblock.append(arr)
                if len(vcfblock) >= self.opt['buff']:
                    f.write(am.get_annotvcf_block(vcfblock))
                    vcfblock = []
            else:
                f.write(am.get_annotvcf_record(arr) + '\n')
            prev_chrom = chrom
            # print(arr[:5])

            if varno > 100:
                break

        if len(vcfblock) > 0:
            f.write(am.get_annotvcf_block(vcfblock))
            vcfblock = []
        '''

    def run(self):
        outvcf = file_util.strip_gzext(self.opt['out'])
        f = open(outvcf, 'w')
        stime0 = time.time()
        total_varno = 0
        am = AnnotMap(self.datastruct, self.datafileinfo)
        vblock = VCFBlockReader(self.opt)
        f.write(vblock.get_header(self.get_version_info() + self.get_annot_header()))
        while(not vblock.eof):
            stime1 = time.time()
            # varno += am.write_annotvcf_with_vcfblock(vblock.get_block(), f)
            varno = am.write_annotvcf_with_vcfblock_tabix(vblock.get_block(), f, self.opt['remove_unannotated_variant'])
            total_varno += varno
            etime1 = time.time()
            elapsed1 = etime1 - stime1
            log = 'processed... ' + str(total_varno) + '/' + str(vblock.total_variant)
            log += " elapsed:" + str(round(elapsed1, 2)) + "s"
            # log += " time:" + str( (elapsed1/varno) * (vblock.total_variant - total_varno))
            print(log)

        etime0 = time.time()
        elapsed0 = etime0 - stime0
        if total_varno > 0:
            print("total:", elapsed0, ", time per variant:", elapsed0 / total_varno,
                  ", time for 4M variants:", str(round(elapsed0 / total_varno * 4000000 / 3600, 3)) + "hr")
            f.close()
        print("Saved " + outvcf)

        if self.opt['out'].endswith('.gz'):
            file_util.save_gzip(file_util.strip_gzext(self.opt['out']))


if __name__ == "__main__":
    pass




VEP_CONSEQUENCE_ORDER = {}
VEP_CONSEQUENCE_ORDER['transcript_ablation'] = 1
VEP_CONSEQUENCE_ORDER['splice_acceptor_variant'] = 2
VEP_CONSEQUENCE_ORDER['splice_donor_variant'] = 3
VEP_CONSEQUENCE_ORDER['stop_gained'] = 4
VEP_CONSEQUENCE_ORDER['frameshift_variant'] = 5
VEP_CONSEQUENCE_ORDER['stop_lost'] = 6
VEP_CONSEQUENCE_ORDER['start_lost'] = 7
VEP_CONSEQUENCE_ORDER['transcript_amplification'] = 8
VEP_CONSEQUENCE_ORDER['inframe_insertion'] = 9
VEP_CONSEQUENCE_ORDER['inframe_deletion'] = 10
VEP_CONSEQUENCE_ORDER['missense_variant'] = 11
VEP_CONSEQUENCE_ORDER['protein_altering_variant'] = 12
VEP_CONSEQUENCE_ORDER['splice_region_variant'] = 13
VEP_CONSEQUENCE_ORDER['incomplete_terminal_codon_variant'] = 14
VEP_CONSEQUENCE_ORDER['start_retained_variant'] = 15
VEP_CONSEQUENCE_ORDER['stop_retained_variant'] = 16
VEP_CONSEQUENCE_ORDER['synonymous_variant'] = 17
VEP_CONSEQUENCE_ORDER['coding_sequence_variant'] = 18
VEP_CONSEQUENCE_ORDER['mature_miRNA_variant'] = 19
VEP_CONSEQUENCE_ORDER['5_prime_UTR_variant'] = 20
VEP_CONSEQUENCE_ORDER['3_prime_UTR_variant'] = 21
VEP_CONSEQUENCE_ORDER['non_coding_transcript_exon_variant'] = 22
VEP_CONSEQUENCE_ORDER['intron_variant'] = 23
VEP_CONSEQUENCE_ORDER['NMD_transcript_variant'] = 24
VEP_CONSEQUENCE_ORDER['non_coding_transcript_variant'] = 25
VEP_CONSEQUENCE_ORDER['upstream_gene_variant'] = 26
VEP_CONSEQUENCE_ORDER['downstream_gene_variant'] = 27
VEP_CONSEQUENCE_ORDER['TFBS_ablation'] = 28
VEP_CONSEQUENCE_ORDER['TFBS_amplification'] = 29
VEP_CONSEQUENCE_ORDER['TF_binding_site_variant'] = 30
VEP_CONSEQUENCE_ORDER['regulatory_region_ablation'] = 31
VEP_CONSEQUENCE_ORDER['regulatory_region_amplification'] = 32
VEP_CONSEQUENCE_ORDER['feature_elongation'] = 33
VEP_CONSEQUENCE_ORDER['regulatory_region_variant'] = 34
VEP_CONSEQUENCE_ORDER['feature_truncation'] = 35
VEP_CONSEQUENCE_ORDER['intergenic_variant'] = 36

VEP_BIOTYPE_ORDER = {}
VEP_BIOTYPE_ORDER['protein_coding'] = 1
VEP_BIOTYPE_ORDER['processed_transcript'] = 5
VEP_BIOTYPE_ORDER['transcribed_processed_pseudogene'] = 10
VEP_BIOTYPE_ORDER['processed_pseudogene'] = 12
VEP_BIOTYPE_ORDER['transcribed_unprocessed_pseudogene'] = 15
VEP_BIOTYPE_ORDER['unprocessed_pseudogene'] = 20
VEP_BIOTYPE_ORDER['miRNA'] = 30
VEP_BIOTYPE_ORDER['snRNA'] = 35
VEP_BIOTYPE_ORDER['lncRNA'] = 40
VEP_BIOTYPE_ORDER['enhancer'] = 50
VEP_BIOTYPE_ORDER['promoter'] = 55
VEP_BIOTYPE_ORDER['promoter_flanking_region'] = 60
VEP_BIOTYPE_ORDER['CTCF_binding_site'] = 70
VEP_BIOTYPE_ORDER['TF_binding_site'] = 80
VEP_BIOTYPE_ORDER['open_chromatin_region'] = 99
VEP_BIOTYPE_ORDER[''] = 100


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from .util import file_util
from .util import proc_util


class MakeDbnsfpTranscript():
    def __init__(self, infile, outfile, ds, opt):
        self.infile = infile
        self.outfile = outfile
        if self.outfile.endswith('.gz'):
            self.outfile2 = outfile[:-3]
        else:
            self.outfile2 = outfile
        self.datastruct = self.get_dbNSFP_Transcipt_struct(ds)

    def get_dbNSFP_Transcipt_struct(self, ds):
        dstruct = file_util.load_json(ds)
        for source in dstruct['source']:
            if source['name'] == "dbNSFP_transcript":
                rst = source
        return rst

    def run(self):
        # print (self.datastruct)

        f = open(self.outfile2, 'w')
        headermap = {}
        i = 0
        for line in file_util.gzopen(self.infile):
            i += 1
            if self.infile.endswith('.gz'):
                line = line.decode('UTF-8')

            arr = line.split('\t')
            arr[-1] = arr[-1].strip()
            if line[0] == '#':
                for k in range(len(arr)):
                    headermap[arr[k].strip()] = k

            enstarr = arr[headermap['Ensembl_transcriptid']].strip().split(';')

            if len(enstarr) > 1:
                # print ("=>", arr)
                for eidx in range(len(enstarr)):
                    cont = []
                    for field in self.datastruct['fields']:
                        arrv1 = arr[headermap[field['name']]].strip().split(';')

                        if len(arrv1) == 1:
                            v1 = arrv1[0]
                        else:
                            # if field['name'] == 'MutationTaster_score':
                            #     print (len(enstarr), field['name'], arrv1, arr[headermap['MutationTaster_pred']], arr[headermap['MutationTaster_model']])
                            v1 = arrv1[eidx].strip()
                        cont.append(v1)
                    # print ("<=", cont)
                    f.write('\t'.join(cont) + '\n')
                # print()
                # break
            else:
                cont = []
                for field in self.datastruct['fields']:
                    v1 = arr[headermap[field['name']]].strip()
                    cont.append(v1)
                f.write('\t'.join(cont) + '\n')

            if i % 10000 == 0:
                print(i, arr[0], arr[1])
                pass
                # break

        f.close()
        if self.outfile.endswith('.gz'):
            proc_util.run_cmd('tabixgz ' + self.outfile2)


if __name__ == "__main__":
    pass




def get_sourcename(s1):
    s1name2 = get_dict_value(s1, 'name2', '')
    s1name = get_dict_value(s1, 'name', '')
    if s1name2 != '':
        s1name = s1name2
    return s1name

def get_dict_value(dict1, key1, default):
    rst = default
    if key1 in dict1.keys():
        rst = dict1[key1]
    return rst


def is_available(field):
    flag = True
    if get_dict_value(field, "is_available", True) == False:
        flag = False
    return flag

def merge_dict(dict1, dict2):
	return(dict2.update(dict1))

def conv_dict_from_arr(arr, header):
	print(len(arr),arr)
	print(len(header), header)
	d = {}
	for i in range(len(arr)):
		d[header[i]] = arr[i]
	return d



import os
import tabix
import time
from .util import file_util
from .util import vcf_util
from .util import proc_util
from .util import struct_util
from .util import region_util
from .model import datasource
from .annot2.annotvcf import VCFAnnotator, VCFVariant
from .model.datastructure import DataSourceListStructure
from .model.datasource import DataSourceList
from . import external_functions
from .renderer import TSIRenderer

REFIDX = 3
ALTIDX = 4
entrezmap = {}
refseqmap = {}



class BlankVCFReader():
    def __init__(self, opt, dslist):
        self.dslist = dslist
        self.opt = opt
        self.region = region_util.Region(opt.region)
        self.tpos = self.region.spos
        self.variantkeylist = []
        self.eof = False
        self.colnames = vcf_util.VCF_COL

    def get_variant(self):
        if len(self.variantkeylist) == 0 and self.tpos < self.region.epos:
            self.load_variantkeys()

        if len(self.variantkeylist) > 0:
            vkey = self.variantkeylist.pop(0)
            vcfline = vkey.replace('_', '\t') + '\t\t\t\t'
            variant = VCFVariant(vcfline, self.colnames, self.opt)
            self.tpos = int(vkey.split('_')[1])
        else:
            variant = None
            self.eof = True
        return variant

    def load_variantkeys(self):
        for r1 in self.region.get_split_region(1000):
            variantkeys = self.dslist.get_variantkeys(r1.chrom, r1.spos, r1.epos)
            poslist = list(variantkeys.keys())
            for pos in sorted(poslist):
                for vkey in variantkeys[pos]:
                    self.variantkeylist.append(vkey)


class DataSourceGenerator(VCFAnnotator):
    def __init__(self, opt):
        self.opt = opt
        self.dslist = DataSourceList(self.opt.ds)
        self.region = region_util.Region(opt.region)
        self.out = None
        self.set_outfile_extension()
        self.vcfreader = BlankVCFReader(self.opt, self.dslist)
        self.tsirenderer = TSIRenderer()

    def set_outfile_extension(self):
        out2 = self.opt.out.replace('.tsi.gz', '.tsi')
        if out2[-4:] != ".tsi":
            out2 += ".tsi"
        self.out = out2

    def get_tsi_header(self):
        header = ["#CHROM", "POS", "ID", "REF", "ALT"]
        info_arr = []
        for s1 in self.dslist.available_source_list:
            info_header = []
            for f1 in s1.available_field_list:
                info_header.append(f1.name2)
            info_arr.append(s1.name + '=' + '|'.join(info_header))
        header.append(';'.join(info_arr))
        return '\t'.join(header)

    def print_i(self, ivar, variant, stime1, stime0):
        if ivar % 500 == 0:
            etime1 = time.time()
            elapsed1 = etime1 - stime1
            elapsed0 = etime1 - stime0
            log = 'processed... ' + str(ivar) 
            log += " elapsed:" + str(round(elapsed1, 2)) + "s " + str(round(elapsed0, 2)) + "s , time for 9G variants:" + str(round(elapsed0 / ivar * 9000000000 / 3600, 3)) + "hr"
            # print(log, ivar, variant)

    def make_single_source_file(self):
        stime0 = time.time()
        self.open_outpointer()
        self.fp.write(self.get_tsi_header() + '\n')

        ivar = 0
        stime1 = time.time()
        while(not self.vcfreader.eof):
            ivar += 1
            variant = self.vcfreader.get_variant()
            if self.vcfreader.eof or variant == None:
                break
            variant.set_annot(self.dslist)
            self.fp.write(self.tsirenderer.render_vcfvariant(variant, False, True))
            self.print_i(ivar, variant, stime1, stime0)
            stime1 = time.time()
        
        self.close_outpointer()
        etime0 = time.time()
        elapsed0 = etime0 - stime0
        print("total:", elapsed0, ", time per variant:", elapsed0 / ivar, ", time for 9G variants:", str(round(elapsed0 / ivar * 9000000000 / 3600, 3)) + "hr")
        # time.sleep(10)
        # proc_util.run_cmd('gz ' + self.out)
        

    def make_single_source_file_with_block(self):
        fp = open(self.opt.out, 'w')
        blockreaders = {}
        for s1 in self.dslist.available_source_list:
            blockreaders[s1.name] = TSVBlockReader(s1)

        tbm = TSVBlockMerger(blockreaders, self.region, int(self.opt.blocksize))
        print_header = True
        while not tbm.eof:
            start = time.time()
            block = tbm.get_block_tsi()
            if print_header:
                fp.write(tbm.get_header())
                print_header = False
            fp.write(block)
            
            end = time.time()
            elapsed = end - start
            log = 'processing.. ' + str(round(elapsed, 3)) + ' sec elapsed. '
            log += str(len(tbm.block.keys())) + ' variants added. '
            log += str(tbm.block_spos) + '~' + str(tbm.block_epos)
            self.log(log)
        fp.close()
        self.log('Saved.. ' + self.out)

    def log(self, msg):
        file_util.fileSave(self.out + '.log', msg + '\n', 'a')
        print(msg)

    def check_datasourcefile(self):
        mt_seq_reader = datasource.MutannoDataSequentialReader(self.out)

        dslist = datasource.DataSourceList()
        dslist.set_datastructure(self.datastruct)

        for mtdata in mt_seq_reader:
            annot = dslist.get_sourcedata_variant(mtdata['CHROM'], mtdata['POS'], mtdata['REF'], mtdata['ALT'])
            
            break
        pass



class TSVBlockReader():

    def __init__(self, datasource):
        self.source = datasource
        self.fileformat = self.source.format
        self.target_colidx = []
        self.header = []
        self.filter_start_with = {}
        self.filter_skip_equal_str = {}
        self.filter_keep_equal_str = {}
        self.delimiter = {}
        self.defaultvalue = {}
        self.field_function = {}
        self.field_function_param = {}
        self.field_names = []
        self.tp = None
        self.chrom = ''
        self.refidx = self.source.ref_column_index
        self.altidx = self.source.alt_column_index
 
    def add_block_bed(self, block, region, sid):
        if self.tp is None:
            sourcefile = self.source.sourcefile2.replace('#CHROM#', region.chrom)
            if file_util.is_exist(sourcefile) and file_util.is_exist(sourcefile + ".tbi"):
                self.tp = tabix.open(sourcefile)
                self.source.set_header(sourcefile)

        if self.tp is not None:
            self.set_target_column()
            for arr in self.tp.querys(region.region_str):
                d = {}
                spos = int(arr[1])
                epos = int(arr[2])
                ref = arr[self.refidx]
                if (spos+1) >= region.spos and (spos+1) <= region.epos:
                    cont = (spos,epos,ref, self.source.name + "=" + "|".join(self.select_field(arr, self.source.header)))
                    try:
                        block[spos]
                    except KeyError:
                        block[spos] = {}
                    try:
                        block[spos]['BED'].append(cont)
                    except KeyError:
                        block[spos]['BED'] = [cont]
        return block

    def set_target_column(self):
        for idx, h1 in enumerate(self.source.header):
            if self.source.format == "bed":
                self.target_colidx = list(range(3, len(self.source.header)))
            elif self.fileformat == "tsi":
                self.target_colidx = [len(self.source.header)-1]
            else:
                for idx, h1 in enumerate(self.source.header):
                    if h1.upper() not in ['CHROM','POS','REF','ALT','VEP']:
                        if idx not in self.target_colidx:
                            self.target_colidx.append(idx)

    def select_field(self, arr, header):
        rst = []
        for idx in self.target_colidx:
            rst.append(vcf_util.encode_value(arr[idx]))
        return rst


    def add_block(self, block, region, sid):
        if self.tp is None:
            sourcefile = self.source.sourcefile2.replace('#CHROM#', region.chrom)
            if file_util.is_exist(sourcefile) and file_util.is_exist(sourcefile + ".tbi"):
                self.tp = tabix.open(sourcefile)
                self.source.set_header(sourcefile)
        if self.tp is not None:
            self.set_target_column()
            try:
                for arr in self.tp.querys(self.source.chrompre + region.region_str):
                    pos = int(arr[1])
                    if pos >= region.spos and pos <= region.epos:
                        refalt = arr[self.refidx] + '_' + arr[self.altidx]
                        try:
                            block[pos]
                        except KeyError:
                            block[pos] = {}
                        try:
                            block[pos][refalt]
                        except KeyError:
                            block[pos][refalt] = {}

                        if self.fileformat == "tsi":
                            cont = arr[-1].replace('&', '~').replace(' ', '%20')
                        else:
                            cont = '|'.join(self.select_field(arr, self.source.header))

                        try:
                            block[pos][refalt][sid].append(cont) 
                        except KeyError:
                            block[pos][refalt][sid] = [cont]
            except tabix.TabixError:
                pass
        
        return block

    def get_header(self):
        rst = []
        for idx in self.target_colidx:
            rst.append(self.source.header[idx])
        return rst



class TSVBlockMerger():
    def __init__(self, block_readers, region, block_size):
        self.block_readers = block_readers
        self.block_size = block_size
        self.region = region  # region object
        self.block_spos = 0
        self.block_epos = 0
        self.block_lpos = 0
        self.block = {}
        self.eof = False

    def read_block(self):
        if self.block_spos == 0:
            self.block_spos = self.region.spos
        else:
            self.block_spos = self.block_epos + 1

        if (self.block_spos + self.block_size) <= self.region.epos:
            self.block_epos = self.block_spos + self.block_size - 1
        else:
            self.block_epos = self.region.epos
        this_range = region_util.Region(self.region.chrom + ':' + str(self.block_spos) + '-' + str(self.block_epos))


        self.block = {}
        for sid in self.block_readers.keys():
            if self.block_readers[sid].fileformat != "bed":
                self.block = self.block_readers[sid].add_block(self.block, this_range, sid)

        for sid in self.block_readers.keys():
            if self.block_readers[sid].fileformat == "bed":
                self.block = self.block_readers[sid].add_block_bed(self.block, this_range, sid)

    def get_block_tsi(self):
        self.read_block()
        if self.block_epos >= self.region.epos:
            self.eof = True

        cont = ''
        poslist = list(self.block.keys())
        for pos in sorted(poslist):
            for refalt in self.block[pos].keys():
                if refalt == "BED":
                    for bedline in self.block[pos]['BED']:
                        spos = bedline[0]
                        epos = bedline[1]
                        ref = bedline[2]
                        infoarr = bedline[3]
                        cont += self.region.chrom + '\t' + str(spos+1) + '\t\t' + ref + '\t\t\t\t' + 'END=' + str(epos) +  '\t' + infoarr + '\n'
                else:
                    infoarr = []
                    for sid in self.block_readers.keys():
                        if sid in self.block[pos][refalt].keys():
                            if self.block[pos][refalt][sid] != "":
                                infoarr.append(sid + '=' + ",".join(self.block[pos][refalt][sid]))
                                
                        cont += self.region.chrom + '\t' + str(pos) + '\t\t' + refalt.replace('_', '\t') + '\t\t\t\t' + ';'.join(infoarr) + '\n'
            

                
            self.block_lpos = pos
        return cont

    def get_header(self):
        cont = ""
        for sid in self.block_readers.keys():
            header = self.block_readers[sid].get_header()
            if cont != "":
                cont += ";"
            cont += sid + '=' + '|'.join(header)
        cont = "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\t" + cont+ "\n"
        return cont



